/*
 *  Copyright (c) 1992, 1993 John E. Davis  (davis@amy.tch.harvard.edu)
 *  All Rights Reserved.
 */
#include <stdio.h>
#include <string.h>
#include "display.h"
#include "sysdep.h"


int Term_Cannot_Insert = 0;
/* 1 if terminal lacks the ability to do into insert mode or into delete
   mode. Currently controlled by S-Lang but later perhaps termcap. */

/* extern int fprintf(FILE *, char *, ...);
   extern int fputs(char *, FILE *); */

char *INS_MODE_STR = "\033[4h";   /* ins mode (im) */
char *EINS_MODE_STR = "\033[4l";  /* end ins mode (ei) */
char *SCROLL_R_STR = "\033[%d;%dr"; /* scroll region */
char *CLS_STR = "\033[2J\033[H";  /* cl termcap STR  for ansi terminals */
char *CLR_BOS_STR = "\033[1J";   /* erase to beg of screen */
char *REV_INDEX_STR = "\033M";     /* sr termcap string */
char *REV_VID_STR = "\033[7m";    /* mr,so termcap string */
char *BOLD_VID_STR = "\033[1m";    /**/
char *UNDL_VID_STR = "\033[4m";    /**/
char *NORM_VID_STR = "\033[m";   /* me,se termcap string */
char *DEL_BOL_STR = "\033[1K\r";  /* cb termcap entry */
char *DEL_EOL_STR = "\033[K";
char *HOME_CURS_STR = "\033[H";   /* ho termcap string */
char *DEL_LINE_STR = "\r\033[K";   /* dl termcap string except cursor at bol */
char *DEL_CHAR_STR = "\033[P";   /* dc */
char *DEL_N_LINES_STR = "\033[%dM";  /* DL */
char *ADD_N_LINES_STR = "\033[%dL";  /* AL */

char *CURS_F_STR = "\033[%dC";    /* RI termcap string */
#define LEN_CURS_F_STR 5

char *CURS_B_STR = "\033[%dD";    /* RI termcap string */
char *CURS_U_STR = "\033[%dA";    /* RI termcap string */
char *CURS_D_STR = "\033[%dB";    /* RI termcap string */
/* cm string has %i%d since termcap numbers columns from 0 */
/* char *CURS_POS_STR = "\033[%d;%df";  ansi-- hor and vert pos */
char *CURS_POS_STR = "\033[%d;%dH";   /* cm termcap string */

int  Output_Rate = 0;	       /* number of chars to output in 1 second */

void send_string_to_term(char *str)
{
   static unsigned long last_time;
   static int total;
   unsigned long now;
   int n = strlen(str);
   
   total += n;
   write(fileno(stdout), str, n);
   if ((Output_Rate > 20) && (total > Output_Rate))
     {
	total = 0;
	if ((now = sys_time()) - last_time <= 1)
	  {
	     sleep((unsigned) 1);
	  }
	last_time = now;
     }
}


void tt_putchar(char ch)
{
   write(fileno(stdout), &ch, 1);
}

void tt_printf(char *fmt, int x, int y)
{
   char buf[255];
   sprintf(buf, fmt, x, y);
   send_string_to_term(buf);
}

void set_scroll_region(int r1, int r2)
{
   tt_printf(SCROLL_R_STR,r1,r2);
}

void reset_scroll_region()
{
    set_scroll_region(1, Screen_Height);
}

void goto_rc(int r, int c)
{
   tt_printf(CURS_POS_STR,r,c);
}

/* void curs_bol()
{
    fputc('\015',stdout);
}

void cursor_forward(int n)
{
    if (n) tt_printf(CURS_F_STR,n, 0);
}

void cursor_backward(int n)
{
    if (n) tt_printf(CURS_B_STR,n, 0);
}

void cursor_up(int n)
{
    tt_printf(CURS_U_STR,n, 0);
}

void cursor_down(int n)
{
    tt_printf(CURS_D_STR,n, 0);
}
*/
void begin_insert()
{
    send_string_to_term(INS_MODE_STR);
}

void end_insert()
{
    send_string_to_term(EINS_MODE_STR);
}

void tt_delete_char()
{
    send_string_to_term(DEL_CHAR_STR);
}

void tt_erase_line()
{
    send_string_to_term(DEL_LINE_STR);
}

void tt_delete_nlines(int n)
{
    if (n) tt_printf(DEL_N_LINES_STR,n, 0);
}

void cls()
{
    send_string_to_term(CLS_STR);
}

void reverse_index(int n)
{
       if (n) tt_printf(ADD_N_LINES_STR,n, 0);
/*
    while(n--)
      send_string_to_term(REV_INDEX_STR); */
}

int beep()
{
   if (!Ignore_Beep) 
     {
	tt_putchar('\007');
     }
   flush_input();
   return(0);
}

void tt_del_eol()
{
    send_string_to_term(DEL_EOL_STR);
}
/*
void tt_open_line(int n)
{
    if (n) tt_printf( ADD_N_LINES_STR, n, 0);
} */

void tt_reverse_video()
{
    send_string_to_term(REV_VID_STR);
}

void tt_normal_video()
{
    send_string_to_term(NORM_VID_STR);
}

void narrow_width()
{
    send_string_to_term("\033[?3l");
}

void wide_width()
{
    send_string_to_term("\033[?3h");
}

void smart_puts(char *neww,char *oldd, int row)
{
   char out[250], curs[20], *mark;
   register char *p, *new = neww, *old = oldd, ch, ch1;
   int ii,max_len,i, curs_set = 0;
   char *new_save;

   
    i = 0;
    ii = 0;
    *curs = 0;
   
   

    /* while they match, go on */
   new_save = new + 1;
   while (((ch = *new++) == *old++) && ch);
   i += new - new_save;
   /* while(ch = *new++, ch1 = *old++, (ch == ch1) && (ch != '\0')) i++; */

    if (ch == '\0')
    /* we are at the end of the new, so delete eond of old line */
      {
	 if ((ch1 = *(old - 1)) == ' ')
	   {
	      while (*old++ == ' ');
	      ch1 = *(old - 1);
	   }

	 if (ch1 == 0) return;

	 goto_rc(row, i + 1);
	 tt_del_eol();
	 return;
      }

    max_len = LEN_CURS_F_STR;

    if (i)
      {
	 sprintf(curs, CURS_POS_STR, row, i + 1);
	 curs_set = 1;
      }

    while(1)
      {
	 ch1 = 0;
	 p = out;
	 *p++ = ch;
	 while (ch = *new++, ch1 = *old++, (ch != ch1) && ch) *p++ = ch;
	 mark = p;
         *p++ = ch;
	 if (ch) while (ch = *new++, ch1 = *old++, (ch == ch1) && ch)
	   {
	      *p++ = ch;
	   }
	 *p = 0;
	 i = p - mark;
	 if (i > max_len)
	   {
	      *mark = 0;
	      if (*curs)
		{
                      send_string_to_term(curs);
                      *curs = 0;
		}
	      if (!curs_set)
		{
		   goto_rc(row, 1);
		   curs_set = 1;
		}

	      send_string_to_term(out);
	      if (ch == '\0')
                  {
                      if (ch1)
                        {
                            old--;
			    while (ch1 = *old++, (ch1 == ' ') && (ch1 != 0));
                        }

		      if (ch1 == 0) return;
		      if (curs_set) sprintf(curs, CURS_F_STR, i);
		      else
			{
			   sprintf(curs, CURS_POS_STR, row, i + 1);
			   curs_set = 1;
			}

		      send_string_to_term(curs);
		      tt_del_eol();
                      return;
                  }

                if (i)
		  {
		     if (curs_set) sprintf(curs, CURS_F_STR, i);
		     else
		       {
			  sprintf(curs, CURS_POS_STR, row, i + 1);
			  curs_set = 1;
		       }
		  }
            }
          else
            {
                if (*curs)
                  {
                      send_string_to_term(curs);
                      *curs = 0;
                  }
                if (!curs_set)
		  {
		     goto_rc(row, 1);
		     curs_set = 1;
		  }
                send_string_to_term(out);
                if (ch == '\0')
                  {
                      if (ch1 == ' ')
                        {
			   while (*old++ == ' ');
			   ch1 = *(old - 1);
                        }

                      if (ch1 != '\0') tt_del_eol();
                      return;
                  }
            }
      }
}

#if 0

void smart_puts(char *new,char *old)
{
    char out[250], ch, ch1, curs[20];
    int ii,max_len,i,mark;

    i = 0;
    ii = 0;
    *curs = 0;

    /* while they match, go on */
    while(ch = *new++, ch1 = *old++, (ch == ch1) && (ch != '\0')) i++;

    if (ch == '\0')
    /* we are at the end of the new, so delete eond of old line */
      {
          if (ch1)
            {
                old--;
                while (ch1 = *old++, (ch1 == ' ') && (ch1 != 0));
            }

	  if (ch1 == 0) return;

          if (i && ch1)  tt_printf(CURS_F_STR, i, 0);
          tt_del_eol();
          return;
      }

    max_len = strlen(CURS_F_STR);
    if (i)
      {
          sprintf(curs, CURS_F_STR, i);
      }

    while(1)
      {
          i = 1; ii = 0;
          ch1 = 0;
          out[ii++] = ch;
          while (ch = *new++, ch1 = *old++, ch != ch1 && ch != '\0') out[ii++] = ch;
          mark = ii;
          out[ii++] = ch;
          if (ch != '\0') while (ch = *new++, ch1 = *old++, ch == ch1 && ch != '\0')
            {
                i++;
                out[ii++] = ch;
            }
          out[ii] = '\0';
          if (i > max_len)
            {
                out[mark] = '\0';
                if (*curs)
                  {
                      send_string_to_term(curs);
                      *curs = 0;
                  }

                send_string_to_term(out);
                if (ch == '\0')
                  {
                      if (ch1)
                        {
                            old--;
			    while (ch1 = *old++, (ch1 == ' ') && (ch1 != 0));
                        }

		      if (ch1 == 0) return;
		      sprintf(curs, CURS_F_STR, i);
		      send_string_to_term(curs);
		      tt_del_eol();
                      return;
                  }

                if (i) sprintf(curs, CURS_F_STR, i);
            }
          else
            {
                if (*curs)
                  {
                      send_string_to_term(curs);
                      *curs = 0;
                  }
                send_string_to_term(out);
                if (ch == '\0')
                  {
                      if (ch1)
                        {
                            old--;
                            while (ch1 = *old++, (ch1 == ' '));
                        }

                      if (ch1 != '\0') tt_del_eol();
                      return;
                  }
            }
      }
}

#endif

