/*
 *  Copyright (c) 1992, 1993 John E. Davis  (davis@amy.tch.harvard.edu)
 *  All Rights Reserved.
 */
#include <stdio.h>
#include <string.h>
#include <limits.h>
#ifndef sequent
#include <stdlib.h>
#endif

#ifdef VMS
#include <file.h>
#endif

#ifdef unix
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#endif

#ifdef msdos
#include <fcntl.h>
#include <io.h>
#include <sys/stat.h>
#endif

#include "buffer.h"
#include "file.h"
#include "misc.h"
#include "sysdep.h"
#include "paste.h"
#include "slang.h"
#include "ins.h"
#include "ledit.h"
#include "mem.h"

#ifdef msdos
#define MAX_LINE_LEN 4096
#else
#define MAX_LINE_LEN 64 * 1024
#endif

int Require_Final_Newline = 0;

/* 0 = read, 1 = write , 2 = append... */
int sys_open(char *file, int acces)
{
   int fp = -1;
   int flags;
   
#ifdef VMS
   char *p, new[255];
#endif
   flags = file_status(file);
   if ((flags < 0) || (flags > 1)) return(fp);
   
   /* on VMS I cheat since I do not want to deal with RMS at this point */
#ifdef VMS
   strcpy(new, file);
   p = new; while (*p) if (*p == ';') *p = 0; else p++;
   
   if (acces == 0) fp = open(file, O_RDONLY, 0);
   else if (acces == 1)
     {
	/* fp = open(new, O_TRUNC, 0); */
	fp = open(new, O_WRONLY | O_TRUNC | O_CREAT, 0, "rat=cr", "rfm=var");
	/* fp = creat(new, 0, "fop=sup", "rat=cr", "rfm=var"); */
	/* fp = open(new, O_WRONLY, 0); */
     }
   
   else if (acces == 2) fp = open(file, O_WRONLY | O_APPEND | O_CREAT);
#else

   switch(acces)
     {
	case 0: flags =  O_RDONLY; break;
	case 1: flags =  O_WRONLY | O_CREAT | O_TRUNC; break;
	case 2: flags =  O_WRONLY | O_CREAT | O_APPEND; break;
	default: return(fp);
     }
#ifdef msdos
   fp = open(file, flags, S_IREAD | S_IWRITE); 
#else
   fp = open(file, flags, 0666); 
#endif
#endif
   return(fp);
}

/* Leaves Point at last line inserted */
/* returns -1 if unable to open file,
           -2 if memory allocation error
           otherwise returns number of lines read */

char *file_type(char *file)
{
    char *p;
    if ((file == (char *) NULL) || (*file == 0)) return( (char *) NULL);

    p = file; while (*p != 0) p++;
    while((p != file) && (*p != '.')) p--;
    if (*p == '.') p++;
    if (p == file) return((char *) NULL); else return(p);
}

void set_file_modes()
{
   char *type;

   if (CBuf == NULL) return;
   CBuf->c_time = sys_time();
   if (CBuf->file[0])
     {
	CBuf->flags |= AUTO_SAVE_BUFFER;
	CBuf->hits = 0;
	type = file_type(CBuf->file);
     }
   else type = (char *) NULL;

   if (type == (char *) NULL) CBuf->modes = NO_MODE;
   else if (lang_run_hooks("mode_hook", type, NULL));

   else if (!strcmp("c", type)) CBuf->modes = C_MODE;
   else if (!strcmp("h", type)) CBuf->modes = C_MODE;
   else if (!strcmp("txt", type)) CBuf->modes = WRAP_MODE;
   else if (!strcmp("tex", type)) CBuf->modes = WRAP_MODE;
   else if (!strcmp("doc", type)) CBuf->modes = WRAP_MODE;
   else CBuf->modes = NO_MODE;
}

#if 0
int read_file_pointer(int fp)
{
   int i = 0, len;                      /* number of lines read in */
   unsigned char *data, buffer[MAX_LINE_LEN], *b1, *bmax;
   register unsigned char *b2;
   unsigned int n, nread;
   int newline;
   
   if (CLine == NULL) make_line(20);
   newline = 0;
   while(1)
     {
	if (0 == (nread = read(fp, buffer, MAX_LINE_LEN)))
	  {
	     if (i == 0)
	       {
		  CLine->next = NULL;
		  CLine->prev = NULL;
		  CBuf->beg = CBuf->end = CLine;
		  return(0);
	       }
	     return(i);
	  }

	b1 = b2 = buffer;
	bmax = b1 + nread;
	while (b2 < bmax)
	  {
	     if (*b2 == '\n')
	       {
		  i++;
		  n = (int) (b2 - b1) + 1;
		  
		  if (newline) (void) make_line(n + 1);
		  len = CLine->len + n;
		  if (len > CLine->space) remake_line(len + 1);
		  data = CLine->data + CLine->len;
		  MEMCPY(data, b1, n);
		  b1 = b2 + 1;
		  CLine->len = len;
		  newline = 1;
	       }
	     b2++;
	  }
	
	if ((n = (int) (bmax - b1)) != 0)
	  {
	     if (newline) (void) make_line(n + 1);
	     len = CLine->len + n;
	     if (len > CLine->space) remake_line(len + 1);
	     data = CLine->data + CLine->len;
	     MEMCPY(data, b1, n);
	     CLine->len = len;
	  }
	if (*(bmax - 1) != '\n') newline = 0;
     }   
}
#endif
int read_file(char *file)
{
   int fp;
   int n, status;

   if ((fp = sys_open(file, 0)) < 0)
     {
	status = file_status(file);
	if (!status) return(-1);  /* file does not exist */
	return(-2); /* exists but not readable */
     }

   n = read_file_pointer(fp);
   close(fp);
   eob();
   if ('\n' == *(CLine->data + Point)) make_line(1);
   return n;
}

int insert_file(char *file)
{
   Buffer *save = CBuf, *tmp;
   int n;

   tmp = make_buffer();
   switch_to_buffer (tmp);

   if ((n = read_file(file)) >= 0)
     {
	switch_to_buffer(save);
	insert_buffer(tmp);
     }
   else switch_to_buffer(save);

   delete_buffer(tmp);
   return(n);
}

/* returns -1 on failure */
int write_region_to_fp(int fp)
{
   int n = 0, pnt, last_pnt, len;
   Line *first, *last;

   if (!check_region(&Number_One)) return(-1);
   last = CLine; last_pnt = Point;

   pop_mark(&Number_One);
   first = CLine; pnt = Point;

   /* first should never be null without hitting last first.  If this
      ever happens, check_region failed. */
   while (first != last)
     {
	len = first->len - pnt;
	if (len != write(fp, (char *) (first->data + pnt), len))
	  {
	     msg_error("Write Failed!");
	  }
	
	/* This goes here inside the loop because it is possible for external
	   events to set error_buffer */
	if (*Error_Buffer) break;
	first = first->next;
	n++;
	pnt = 0;
     }

   if (last_pnt != 0)
     {
	len = last_pnt - pnt;
	if (len != write(fp, (char *) (last->data + pnt), len))
	  {
	     msg_error("Write Failed!");
	  }
	n++;
     }
#ifndef VMS
   if ((Require_Final_Newline) && (CBuf->end == last))
     {
	eob(); if (Point) write(fp, "\n", 1);
     }
#endif
   pop_spot();
   if (*Error_Buffer) return(-1);
   return(n);
}

/* write current buffer to open file pointer. Return number of lines */

int write_region(char *file)
{
   int fp;
   int n;
   char msg[255];

   if ((fp = sys_open(file, 1)) < 0)
     {
	sprintf(msg, "Unable to open %s for writing.", file);
	msg_error(msg);
	return(-1);
     }
   n = write_region_to_fp(fp);
   close(fp);

   return(n);
}



/* returns -1 on failure and number of lines on success */

int write_file(char *file)
{
   int n;
   push_spot();
   bob();
   push_mark();
   eob();
   if (-1 == (n = write_region(file))) pop_mark(&Number_Zero);
   pop_spot();
   return(n);
}

int append_to_file(char *file)
{
   int fp;
   int n;

   if ((fp = sys_open(file, 2)) < 0) return(-1);
   n = write_region_to_fp(fp);
   close(fp);
   check_buffers();
   return(n);
}

int make_autosave_filename(char *save, char *dir, char *file)
{
   char *s;
   int dat;
   
   if (*file == 0) return(0);
   
   
   if (lang_run_hooks("make_autosave_filename", dir, file) && !Lang_Error)
     {
	if (!lang_pop_string(&s, &dat)) return(0);
	strncpy(save, s, 255);
	save[255] = 0;
	if (dat) FREE(s);
     }
   else
     {
#ifdef unix
	sprintf(save, "%s#%s#", dir, file);
#else
#ifdef msdos
	sprintf(save, "%s#%s", dir, file);
#else
	sprintf(save, "%s_$%s;1", dir, file);
#endif
#endif
     }
   return(1);
}

int write_file_with_backup(char *dir, char *file)
{
   char *old;
   char new[255]; char save[255];
   int n, bu, do_free = 0;
   int mode, do_mode;
#ifndef VMS
   int ok = 0;
#endif

   if (*file == 0) return(-1);

   sprintf(new, "%s%s", dir, file);

   do_mode = sys_chmod(new, 0, &mode);
   if ((do_mode < 0) ||  (do_mode > 1)) return(-1);
#ifndef VMS
   bu = 1;
   if (lang_run_hooks("make_backup_filename", dir, file))
     {
	if (((bu = lang_pop_string(&old, &do_free)) != 0) && (*old))
	  {
	     unlink(old);
	     ok = rename(new, old);
	  }
     }
#endif
   
   if (-1 != (n = write_file(new)))
     {
	if (make_autosave_filename(save, dir, file))
	  {
	     sys_delete_file(save);
	  }
	     
	if (do_mode) /* must be an existing file, so preserve mode */
	  {
	     sys_chmod (new, 1, &mode);
	  }
	CBuf->c_time = sys_time();
     }
#ifndef VMS
   else if (ok && bu && *old) rename(old, new); /* error -- put it back */
   if (do_free && bu) FREE(old);
#endif
   return(n);
}

/* warning-- this saves on the narrowed part of buffer.
   Here, I widen first.  I need a save_restriction type of thing because
   I do not narrow back.
   */
void auto_save_buffer(Buffer *b)
{
    char tmp[255];
    Buffer *old_buf;

   if ((b == NULL) || (b->narrow != NULL)) return;
   old_buf = CBuf;
   CBuf = b;

    if ((b->flags & BUFFER_TRASHED) && (b->flags & AUTO_SAVE_BUFFER))
      {
	  if (make_autosave_filename(tmp, b->dir, b->file))
	    {
	       flush_message("autosaving..."); 
	       sys_delete_file(tmp);
	       write_file(tmp);
	       message("autosaving...done");
	       b->hits = 0;
	    }
      }

    CBuf = old_buf;
}

void auto_save_all()
{
    Buffer *b;

   if (NULL == (b = CBuf)) return;
   do
     {
	while (b->narrow != NULL) widen_buffer(b);
	if (*b->file != 0) auto_save_buffer(b);
	b = b->next;
     }
   while (b != CBuf);
}

void visit_file(char *dir, char *file)
{
   if (NULL == find_buffer(file)) strcpy(CBuf->name, file);
   strcpy(CBuf->dir, dir);
   CBuf->c_time = sys_time();
   strcpy(CBuf->file, file);
   check_buffers();
}

void fixup_dir(char *dir)
{
#ifndef VMS
#ifdef msdos
   if (dir[strlen(dir) - 1] != '\\') strcat(dir,"\\");
#else
   if (dir[strlen(dir) - 1] != '/') strcat(dir,"/");
#endif
#endif
}

char *dir_file_merge(char *dir, char *file)
/*
;;  returns result of merging dir with file. If file is empty, dir is
;;  considered to be whole file.
*/
{
   char name[512];

   strcpy (name, dir);
   if ((file != NULL) && *file)
     {
	fixup_dir(name);
	strcat(name, file);
     }
   return expand_filename(name);
}

int file_status(char *file)
/*
;;  Returns a coded integer about file.  If the file does not exist, 0 is
;;  returned.  Meaning:
;;
;;     2 file is a directory
;;     1 file exists
;;     0 file does not exist.
;;    -1 no access.
;;    -2 path invalid
;;    -3 unknown error
*/
{
   int mode = 0;
   return sys_chmod(file, 0, &mode);
}

int file_changed_on_disk(char *file)
{
   unsigned long t;
   Buffer *buf;
   if (NULL == (buf = find_file_buffer(file))) return(0);
   t = sys_file_mod_time(file);
   return(t > buf->c_time);
}

int file_time_cmp(char *file1, char *file2)
{
   unsigned long t1, t2;
   
   t1 = sys_file_mod_time(file1);
   t2 = sys_file_mod_time(file2);

   if (t1 > t2) return(1); else if (t1 == t2) return 0; else return(-1);
}

void auto_save(void)
/*
;;  autosave current buffer if marked for autosave.  To autosave all buffers
;;  use autosvall
;;
*/
{
   auto_save_buffer(CBuf);
}

void check_buffer(Buffer *b)
{
   if ((*b->file != 0)
       && file_changed_on_disk(dir_file_merge(b->dir, b->file)))
     {
	b->flags |= FILE_MODIFIED;
     }
   else b->flags &= ~FILE_MODIFIED;
}

void check_buffers()
{
   Buffer *b = CBuf;

   do
     {
	check_buffer(b);
	b = b->next;
     }
   while (b != CBuf);
}

#ifdef msdos
void ibm_set_file_trans(int *mode)
{
   if (*mode) _fmode = O_BINARY; else _fmode = O_TEXT;
}   
#endif

/*
char *my_scan(char *p, char ach, int n)
{
   if (n == 0) return p;
   asm {
      mov ax, di
      mov al, ach
      les di, p
      mov cx, n
      cld
      repne scasb
      inc cx
      sub n, cx
      mov di, ax
   }
   p = p + n;
   if (*p != ach) p++;
   return p;
}
*/
char *vgets(VFILE *vp, unsigned int *num)
{
   register char *bp, *bp1;
   char *bmax, *bpmax;
   char *new;
   int fd = vp->fd;
   unsigned int n, max;
#ifndef msdos
   unsigned int n2;
#endif
   int doread = 0;
   n = vp->size;

   *num = 0;
   if (NULL == vp->buf)
     {
	if (NULL == (new = (char *) MALLOC(n))) 
	  {
	     msg_error("Unable to malloc space.");
	     return(NULL);
	  }
	
	vp->bp = vp->buf = new;
	vp->bmax = new + n;
	doread = 1;
     }

   bp = vp->bp;
   if ((vp->eof != NULL) && (bp >= vp->eof)) return (NULL);
   bp1 = vp->buf;
   bmax = vp->bmax;
   
   while (1)
     {	
	if (doread)
	  {
	     max = (int) (vp->bmax - bp);
	     while ((n = read(fd, bp, max)) != 0)
	       {
		  max -= n;
		  bp += n;
	       }
	     if (max) vp->eof = bp;
	     if (bp == bp1) return(NULL);
	     bp = bp1;
	  }
	else bp1 = bp;
	   
	/* extract a line */
	if (vp->eof != NULL) bmax = vp->eof;
	
	n = (int) (bmax - bp);
#ifdef msdos
	if (n)
	  {
	     bpmax = bp;
	     asm  {
		mov bx, di
		mov al, 10
		les di, bpmax
		mov cx, n
		cld
		repne scasb
		inc cx
		sub n, cx
		mov di, bx
	     }
	     bp += n;
	     if (*bp != '\n') bp++;
	  }
	
	if (bp < bmax)
	  {
	     vp->bp = ++bp;
	     *num = (unsigned int) (bp - bp1);
	     return bp1;
	  }
#else
	n2 = n % 4;
	bpmax = bp + (n - 4);
	
	while(bp <= bpmax)
	  {
	     if ((*bp++ == '\n')
		 || (*bp++ == '\n')
		 || (*bp++ == '\n')
		 || (*bp++ == '\n'))
	       {
		  vp->bp = bp;
		  *num = (unsigned int) (bp - bp1);
		  return bp1;
	       }
	  }
	while(n2--)
	  {
	     if (*bp++ == '\n')
	       {
		  vp->bp = bp;
		  *num = (unsigned int) (bp - bp1);
		  return bp1;
	       }
	  }
#endif	
	if (vp->eof != NULL)
	  {
	     /* *bp = 0; */
	     *num = (unsigned int) (bp - bp1);
	     vp->bp = bp;
	     return(bp1);
	  }
	
	doread = 1;

	bp = bp1;
	bp1 = vp->buf;
	if (bp != bp1)
	  {
	     /* shift to beginning */
	     while (bp < bmax) *bp1++ = *bp++;
	     bp = bp1;
	     bp1 = vp->buf;
	  }
	else 
	  {
	     bp = bmax;
	     vp->bmax += 2 * (int) (vp->bmax - vp->buf);
	     new = (char *) REALLOC(vp->buf, vp->bmax - vp->buf);
	     if (new == NULL) 
	       {
		  msg_error("Realloc Error.");
		  return(NULL);
	       }
	     
	     bp = new + (int) (bmax - vp->buf);
	     bmax = vp->bmax = new + (int) (vp->bmax - vp->buf);
	     bp1 = vp->buf = new;
	  }
     }
}



VFILE *vstream(int fd, int size)
{
   VFILE *v;
   
   if (NULL == (v = (VFILE *) MALLOC(sizeof(VFILE)))) return(NULL);
   v->bmax = v->bp = v->buf = NULL;
   v->fd = fd;
   v->eof = NULL;
   v->size = size;
   return v;
}

   
VFILE *vopen(char *file, int size)
{
   int fd;
   
   if ((fd = open(file, O_RDONLY, 0)) < 0) return(NULL);
   return vstream(fd, size);
}

void vclose(VFILE *v)
{
   close(v->fd);
   if (v->buf != NULL) FREE(v->buf);
   FREE(v);
}


int read_file_pointer(int fp)
{
   int n = 1;
   unsigned int num;
   unsigned char *vbuf;
   VFILE *vp;
   
   if ((vp = vstream(fp, MAX_LINE_LEN)) == NULL) return(-1);
   
   if (NULL == (vbuf = (unsigned char *) vgets(vp, &num))) return(0);
   if (CLine->space < num)
     {
	remake_line(CLine->len + num);
     }
   MEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
   CLine->len = num;
   
   while(NULL != (vbuf = (unsigned char *) vgets(vp, &num)))
     {
	n++;
	make_line(num);
	MEMCPY((char *) CLine->data, (char *) vbuf, (int) num);
	CLine->len = num;
     }
   if (vp->buf != NULL) FREE(vp->buf);
   FREE(vp);
   return n;
}


#ifdef msdos
unsigned long sys_file_mod_time(char *file)
{
   struct stat buf;
   /* int fd; */
   /* if (NULL == (fd = sys_open(file, 0))) return(0); */
   if (stat(file, &buf) != 0) return(0);
   /* close(fd); */
   return ((unsigned long) buf.st_mtime);
}
#endif
