/*
 *  Copyright (c) 1992, 1993 John E. Davis  (davis@amy.tch.harvard.edu)
 *  All Rights Reserved.
 */
#include <stdio.h>
#include <string.h>
#include <setjmp.h>
#ifdef unix
#include <sys/file.h>
#endif
#ifdef msdos
#include <dir.h>
#include <io.h>
extern unsigned _stklen = 20000U;
#endif

#include "file.h"
#include "buffer.h"
#include "display.h"
#include "sysdep.h"
#include "sig.h"
#include "misc.h"
#include "keymap.h"
#include "screen.h"
#include "ledit.h"
#include "slang.h"
#include "search.h"
#include "text.h"

typedef struct
{
   jmp_buf b;
} jmp_buf_struct;

extern jmp_buf_struct Jump_Buffer, *Jump_Buffer_Ptr;

int Batch = 0;

char *get_cwd(void)
{
   static char cwd[255];

   if (*cwd ==  0)
     {
#ifdef sequent
 	if(!getwd(cwd)) return "";
#else
  	getcwd(cwd, 254);
#endif
#ifndef VMS
	fixup_dir(cwd);
#endif
     }
   return(cwd);
}

#if 0
void load_init_file(void)
{
   FILE *fp;
#ifdef unix
   char *file = ".jedrc";
#else
   char *file = "jed.rc";
#endif
   char initfile[255], *home;

   home = getenv("JED_HOME");
#ifndef VMS
   if ((home != NULL) || (NULL != (home = getenv("HOME"))))
     {
	strcpy(initfile, home);
	fixup_dir(initfile);
     }
   else initfile[0] = 0;
#else
   if (home == NULL) home = "SYS$LOGIN:";
   strcpy(initfile, home);
#endif

   strcat(initfile, file);
   if ((fp = fopen(initfile,"r")) == NULL) return;
   fclose(fp);
   lang_load_file(initfile);
}
#endif 

int Main_Argc;
char **Main_Argv;

int main(int argc, char **argv)
{
   char *cmd_hook = "command_line_hook";
   int i,j, fd;

   
   /* Change case Lut.  Do it now so it can be used */
   for (i = 0; i < 256; i++) 
     {
	Chg_UCase_Lut[i] = i;
	Chg_LCase_Lut[i] = i;
     }
   
   for (i = 'A'; i <= 'Z'; i++) 
     {
	j = i + 32;
	Chg_UCase_Lut[j] = i;
	Chg_LCase_Lut[i] = j;
     }
#ifdef msdos
   /* Initialize for DOS code page 437. */
   Chg_UCase_Lut[135] = 128; Chg_LCase_Lut[128] = 135;
   Chg_UCase_Lut[132] = 142; Chg_LCase_Lut[142] = 132;
   Chg_UCase_Lut[134] = 143; Chg_LCase_Lut[143] = 134;
   Chg_UCase_Lut[130] = 144; Chg_LCase_Lut[144] = 130;
   Chg_UCase_Lut[145] = 146; Chg_LCase_Lut[146] = 145;
   Chg_UCase_Lut[148] = 153; Chg_LCase_Lut[153] = 148;
   Chg_UCase_Lut[129] = 154; Chg_LCase_Lut[154] = 129;
   Chg_UCase_Lut[164] = 165; Chg_LCase_Lut[165] = 164;
   /* Note that Slang uses decimal  instead  of Octal.  Octal below, 
      decimal above */
   define_word("0-9a-zA-Z\200-\232\240-\245\341-\353");
#else
   /* ISO Latin */
   for (i = 192; i <= 221; i++) 
     {
	j = i + 32;
	Chg_UCase_Lut[j] = i;
	Chg_LCase_Lut[i] = j;
     }
   Chg_UCase_Lut[215] = 215; Chg_LCase_Lut[215] = 215;
   Chg_UCase_Lut[223] = 223; Chg_LCase_Lut[223] = 223;
   Chg_UCase_Lut[247] = 247; Chg_LCase_Lut[247] = 247;
   Chg_UCase_Lut[255] = 255; Chg_LCase_Lut[255] = 255;

   define_word("0-9a-zA-Z\277-\326\330-\336\340-\366\370-\376");
#endif
   
   if ((argc > 1) && (!strcmp(argv[1], "-batch"))) 
     {
	Batch = 1;
	Lang_Traceback = 1;
     }

#ifdef unix
   if (Batch) fd = 2;
   else 
     {
	if ((fd = open("/dev/tty", O_RDWR)) < 0)
	  {
	     exit_error("Unable to open /dev/tty.");
	  }
   
	dup2(fd, 2);  /* tty uses 2 as the descriptor */
     }
   
#else
   fd = 2;			       /* 2 is stderr, assume it is ok */
#endif
   
   init_tty();
   init_display(1);         /* sets up virtual screen */

   Jump_Buffer_Ptr = &Jump_Buffer;
   
   /* incase something goes wrong before we even get started... */
   if (setjmp(Jump_Buffer_Ptr->b) != 0)
     {
	/* hmm.... just exit I guess */
	exit_error("Fatal Error!");
     }

   if (NULL == (CBuf = make_buffer()))
     {
	exit_error("Allocation Failure.");
     }
   CLine = NULL;
   strcpy(CBuf->name, "*scratch*");
   *CBuf->file = 0;
   strcpy(CBuf->dir, get_cwd());

   /* This order here is crucial! */
   init_keymaps();
   init_minibuffer();

   set_file_modes();
   
   /* what if someone pipes something to jed, allow it */
   
   if (!isatty(0))   /* 1 if stdin is a terminal, 0 otherwise */
     {
	set_buffer("*stdin*");
	read_file_pointer(fileno(stdin));
	bob();
	fclose(stdin);
	dup2(fd, 0);	
     }
   if (CLine == NULL) make_line(25);
   Point = 0;
   window_buffer(CBuf);
   if (!Batch) cls();
#ifndef VMS
#ifndef msdos
   init_signals();
#endif
#endif

   lang_load_file("site.sl");
   
   /* set up command line args */
   Main_Argc = argc;
   Main_Argv = argv;
   
   if (lang_is_defined(cmd_hook))
     {
	/* I do not push "-batch" string! */
	/* argc--;
	for (i = argc; i > Batch; i--) lang_push_string(argv[i]);
	lang_push_string(argv[0]);
	lang_push_integer(argc - Batch); */
	lang_run_hooks(cmd_hook, NULL, NULL);
     }
   else if (argc > 1 + Batch) find_file_in_window(argv[1 + Batch]);   
	

   /* after we have possible loaded key definitions, we can fix up
      the minibuffer map. This way user definitions are used. */

   if (NULL == (Mini_Map = create_keymap("Mini_Map"))) exit_error("Fatal.");
   undefine_key("^M", Mini_Map);
   define_key1("^M", (VOID *) exit_minibuffer, F_INTRINSIC, Mini_Map);
   define_key1("^I", (VOID *) mini_complete, F_INTRINSIC, Mini_Map);
   define_key1(" ", (VOID *) mini_complete, F_INTRINSIC, Mini_Map);

   /* edit_loop */
   if (!Batch) jed();  /* never returns */
   exit_error("Batch End with no exit!");
   return(-1);
}
