/*
 *  Copyright (c) 1992, 1993 John E. Davis  (davis@amy.tch.harvard.edu)
 *  All Rights Reserved.
 */
#define JED_VERSION "Jed <0.91.i> "

#include <stdio.h>
#include <string.h>
#ifdef msdos
#include <dos.h>
extern int Cursor_Row;
extern void init_video(void);
#endif

#include "malloc.h"
#include "buffer.h"
#include "screen.h"
#include "window.h"
#include "paste.h"

#include "ins.h"
#include "ledit.h"
#include "display.h"
#include "sysdep.h"
#include "misc.h"
#include "vterm.h"
#include "slang.h"
#include "mem.h"

#define fputs(s,f) write(fileno(f), s, strlen(s))


Screen_Type Screen[80];

int Screen_Row = 1;
int Screen_Col = 1;
int Cursor_Motion;    /* indicates cursor movement only -1 ^ v +1 < > */
int Scroll_Region_1;
int Scroll_Region_2;
int Scroll_Lines;
int User_Prefers_Line_Numbers = 0;

int Goal_Column;
int Display_Eight_Bit = 0;

int JED_CSI = -1;


static Line Eob_Line =
{
   NULL, NULL, (unsigned char *) "[EOB]", 5, 5
};

int Want_Eob = 0;


void blank_line(int row)
{
   int n = Window->width;
   char *p = Screen[row].old;
   MEMSET(p, ' ', n);
}

void update_screen_txt(int row)
{
    char *tmp;
    tmp = Screen[row].new;
    Screen[row].new = Screen[row].old;
    Screen[row].old = tmp;
}
static unsigned char char_width[256];
static int dis8bit = -100;
static int j_csi = -1;
static int Selective_Display = -0x7FFF;

#define FIX_CHAR_WIDTH \
  if ((Display_Eight_Bit != dis8bit) || (JED_CSI != j_csi)\
      || (Selective_Display != Buffer_Local.sd)) fix_char_width()

	 

int output(unsigned char *line, int len, int row, register int max_col, register int column_offset)
{
   register int count;
   register unsigned char *p, *b, *outp, *pmax, ch;
   unsigned char expb[80];
   unsigned char *out, *mark;
   int tab, tabsize, visible, i, mini, w;
   int done = 0;

   outp = out = (unsigned char *) Screen[row - 1].new;
   tab = Buffer_Local.tab;
   count = 0;
   visible = 0;

   mini = 1;
   while(mini >= 0)
     {
	/* deal with mini_buffer prompt */
	if (mini && (IS_MINIBUFFER) && Mini_Info.prompt_len)
	  {
	     p = (unsigned char *) Mini_Info.prompt;
	     pmax = p + Mini_Info.prompt_len;
	     mini = 0;
	  }
	else
	  {
	     p = line;
	     pmax = p + len;
	     mini = -1;
	  }

	while(p < pmax)
	  {
	     ch = *p++;
	     b = expb;
	     /* expand char to display form */
	     
	     w = *(char_width + (unsigned int) ch);
	     if (w == 1)
	       {
		  /* try to optimize this by putting this here */
		  
		  count++;
		  if ((count < max_col) && (count >= column_offset))
		    {
		       *outp++ = ch;
		    }
		  else if (count >= max_col)
		    {
		       done = 1;
		       p = pmax;
		    }
		  continue;
	       }
	     else if (w == 3)
	       {
		  if (ch == '\r')
		    {
		       *b++ = '.'; *b++ = '.'; *b++ = '.';
		       done = -1;
		       p = pmax;
		    }
		  else
		    {
		       *b++ = '~';
		       *b++ = '^';
		       ch &= 0x7F;
		       *b++ = (ch == 127) ? '?' : ch  + 'A' - 1;
		    }
		  
	       }
	     else if (w == 2)
	       {
		  if ((ch == '\t') && (tab > 0))
		    {
		       tabsize = tab * (count / tab + 1) - count;
		       i = 0; while (i++ < tabsize) *b++ = ' ';
		       w = tabsize;
		    }
		  else if (ch & 0x80)
		    {
		       *b++ = '~'; *b++ = ch & 0x7F;
		    }
		  else
		    {
		       *b++ = '^';
		       *b++ = (ch == 127) ? '?' : ch  + 'A' - 1;
		    }
	       
	       }
	     else		       /* w = 0 */
	       {
		  ch = *p++;	       /* skip next char */
		  continue;
	       }

	     b = expb;
	     while (w--)
	       {
		  count++;
		  if (count >= max_col)
		    {
		       done = 1;
		       p = pmax;
		    }
		  
		  else if (count >= column_offset)
		    {
		       *outp++ = *b;
		    }
		  b++;
	       }
	  }
     }


   p = out;
   while (p < outp) 
     {
	if (*p++ != ' ') 
	  {
	     visible++;
	     break;
	  }
     }
   
   if (outp == out)
     {
	*outp++ = '$';
	visible = 1;
     }
   mark = outp;
   p = outp;  pmax = out + Window->width;
#ifdef msdos
   asm mov cx, word ptr pmax
   asm sub cx, word ptr p
   asm mov al, ' '
   asm les di, dword ptr p
   asm cld
   asm rep stosb
   /* Note that p is not updated here! */
#else
   MEMSET(p, ' ', pmax - p);
   /* while (p < pmax) {*p++ = ' '; *p++ = ' ';} */
#endif
   if (done == 1)
     {
	mark = NULL;
	out[Window->width - 1] = '$';
	visible += 1;
     }
   out[Window->width] = '\0';

   if (visible)
     {
#ifdef msdos
	(void) mark;
	smart_puts((char *) out,Screen[row-1].old, row);
#else
	if (mark != NULL) *mark = 0;
	smart_puts((char *)out,Screen[row-1].old, row);
	if (mark != NULL) *mark = ' ';
#endif
	Screen[row-1].n = visible;
     }

   /* else take care of it in calling routine */
   return(visible);
}

void display_line(Line *line, int row)
{
    int len;
   
   if (line != NULL)
     {
	len = line->len;
	if (len && (row != Screen_Height) 
	    && (line->data[len - 1] == '\n')) len--;

	if ((len > 0) || (Window->column > 1)
	    || ((row == Screen_Height) && Mini_Info.prompt_len))
	  {
	     len = output(line->data, len, row,
			  Window->column + Window->width - 1, Window->column);
	  }
     }
   else len = 0;
   
   if (len <= 0)
     {
	if (Screen[row-1].n)
	  {
/* #ifndef msdos 
	     if (Screen_Row + 1 == row) fputs("\r\n", stdout);
	     else 
	       {
#endif
	     goto_rc(row,1);
#ifndef msdos
	       }
#endif */
	     goto_rc(row, 1);
	     Screen_Row = row;
	     tt_del_eol();
	  }
	blank_line(row - 1);
	Screen[row - 1].n = 0;
     }
   Screen[row - 1].line = line;
   Screen[row - 1].flags = 0;
   if (len > 0) update_screen_txt(row-1);
}

void open_scroll(void)
{
    Scroll_Region_1 = 1;
    Scroll_Region_2 = Screen_Height;
    Scroll_Lines = 0;
}

void do_scroll_up(int n)
{
   goto_rc(1,1); 
   tt_delete_nlines(n);
}

void do_scroll_down(int n)
{
   goto_rc(1,1); 
   reverse_index(n);
}

void execute_scroll(void)
{
   if (Scroll_Lines > 0)
     {
	set_scroll_region(Scroll_Region_1, Scroll_Region_2);
	do_scroll_up(Scroll_Lines);
     }
   else if (Scroll_Lines < 0)
     {
	set_scroll_region(Scroll_Region_1, Scroll_Region_2);
	do_scroll_down(-Scroll_Lines);
     }
   Scroll_Lines = 0;
}

void queue_scroll(int n, int r1, int r2)
{
   if ((r1 != Scroll_Region_1) || (r2 != Scroll_Region_2))
     {
/*	if ((n == -1) && (Scroll_Region_1 == r1 + Scroll_Lines)
	       && (Scroll_Region_2 == r2))  */
	if ((n == -1) && (Scroll_Region_1 == r1 + Scroll_Lines)
	    && (Scroll_Region_2 == r2))  Scroll_Lines--;
	else if ((n == -1) && (Scroll_Region_1 == r1 + Scroll_Lines)
		 && (Scroll_Region_2 == r2 - 1))
	  {
	     Scroll_Region_2 = r2;
	     Scroll_Lines--;
	  }

	else
	  {
	     execute_scroll();
	     Scroll_Region_1 = r1;
	     Scroll_Region_2 = r2;
	     Scroll_Lines = n;
	  }
     }
   else Scroll_Lines += n;
}

void close_scroll(void)
{
    if (Scroll_Lines) execute_scroll();
    if ((Scroll_Region_1 != 1) || (Scroll_Region_2 != Screen_Height))
      reset_scroll_region();
}

/* Here a scrolling region (t,b) is used to scroll line t + n to t.  All
   that is assumed is that  the line at t + n exists! */

int scroll_up(int n, int t, int b)
{
    int i, necess;

    if (n == 0) return(0);

    /* t = Window->top - 1 + t;
       b = Window->top - 1 + b; */

    /* check to see if this is really necessary */
    necess = 0;
    for (i = t - 1; i < b; i++)
      {
          if (Screen[i].n)
            {
                necess = 1;
                break;
            }
      }

    if (!necess) return(0);

    queue_scroll(n, t, b);

    vscroll_up(t,b,n);

    return(necess);
}

/* Here a scrolling region (t,b) is used to scroll line t + n to t. */

void scroll_down(int n, int t, int b)
{
    if (n == 0) return;

    /* t = Window->top - 1 + t;
       b = Window->top - 1 + b; */
    set_scroll_region(t,b);
    reverse_index(n);
    reset_scroll_region();

    vscroll_down(t,b,n);
}

int update_insert_line(int r1, Line *line)
{
    int i, r2, r, necess;
    Line *bot;

    /* normalize r1: */
    /* r1 = Window->top - 1 + r1; */

    /* find the first one that is blank to delimit the region so that the
       loss will be minimal */

    r = r1;   /* not r1 + 1 as obvious (but naive) as it seems */
    r2 = Window->rows + Window->top - 1;
    while (r < r2)
      {
	 bot = Screen[r - 1].line;
	 if ((bot == NULL) || (bot == &Eob_Line)) break;
          r++;
      }

    if ((r1 != r2) && (r == r2))
    /* we may have failed so check the bottom up so we don't push linesoff. */
     {
	bot = line;
	for (r = r1 + 1; r <= r2; r++)
	  {
	     bot = bot->next;
	     if (bot == NULL) break;
	     if (Screen[r-1].line == bot) break;
	  }
	if ((bot != NULL) && (r <= r2)) r--;
	if (r > r2) r = r2;
     }

   if (r < r1) r = r1;
   r2 = r;

    /* check to see if we gain by doing this */
    necess = 0;
    for (i = r1 - 1; i < r2; i++)
      {
          if (Screen[i].n)
            {
                necess = 1;
                break;
            }
      }
    if (r1 == r2) return(0);
    if (!necess) return(0);

    queue_scroll(-1, r1, r2);

    vscroll_down(r1, r2, 1);
    return(1);
}

Line *find_top()
{
    int n, i;
    Line *line, *next, *prev, *this;

    if (Window->rows == 1) return(CLine);
    /* Check the top window line.  Chances are that one of the lines
     above CLine will match in usual situations */
    line = Screen[Window->top - 1].line;
    prev = CLine;

    if (line != NULL) for (i = 0; i < Window->rows; i++)
      {
          if (prev == line) return(line);
          this = prev->prev;
          if (this == NULL) break;
          else prev = this;
      }

    /*  That was the obvious choice now try some others */
    /* try bottom */
    line = Screen[Window->rows + Window->top - 2].line;
    next = CLine;

    if (line != NULL) for (i = 0; i < Window->rows; i++)
      {
          if (next == line)
            {
                line = CLine;
                while(i < Window->rows - 1)
                  {
                      prev = line->prev;
                      if (prev == NULL) return(line);
                      line = prev;
                      i++;
                  }
                return(line);
            }
          next = next->next;
          if (next == NULL) break;
      }
    /* try to find CLine somewhere in the window */

    next = CLine->next;
    prev = CLine->prev;

    for (i = Window->top - 1; i < Window->top - 1 + Window->rows; i++)
      {
         this = Screen[i].line;
         if (this == CLine)
           {
              line = CLine;
              while(i > Window->top - 1)
                {
		    i--;
		    line = line->prev;
		    if (line == NULL) return(CLine);
                }

              return(line);
           }
         else if ((this == prev) && (this != NULL))
           {
               i--;
               while(i-- > Window->top - 1)
                 {
                     if (prev->prev == NULL) return(prev);
                     prev = prev->prev;
                 }
               return(prev);
           }
         else if ((this == next) && (this != NULL))
           {
               i++;
               while(i-- > Window->top - 1)
                 {
                     if (next->prev == NULL) return(next);
                     next = next->prev;
                 }
               return(next);
           }

      }

    /* not found so check neighbors */
    line = CLine->next;
    if ((line != NULL) && (line == Screen[Window->top - 1].line)) return(CLine);

    n = Window->rows - 1;
    line = CLine->prev;
    if ((line == NULL) || (line != Screen[n + Window->top - 1].line))
      {
         n = n / 2;
      }

    line = prev = CLine;
    while(n--)
      {
          line = line->prev;
          if (line == NULL) return(prev);
          prev = line;
      }

    return(line);
}

void do_scroll(register Line *top)
{
   Line *bot, *middle1, *s;
   register Line *sline;
   int r1, r, r2, match, overwrite = 1, scroll_region_set = 0, srs;
   int n1, n2;
   
    r2 = Window->top + Window->rows - 1;
    r1 = Window->top;

    open_scroll();
    if (top != NULL) bot = top->next; else bot = NULL;

#if 0
   s = Screen[r2 - 2].line;   /* was - 1 */
   if (bot != NULL) for (r = r1 + 1; r <= r2; r++)
     {
	if (bot == s)
	  {
	     overwrite = 0;
	     break;
	  }
	if (bot == NULL) break;
	bot = bot->next;
     }
#endif
   
    middle1 = NULL;
#if 0
    if (overwrite)
      {
#endif
          r = 1;
          bot = top;
          n1 = r1 + 3;
          while (r++ < Window->rows)
            {
	       if ((r >= n1) && (middle1 == NULL) && (bot != NULL)
		   && (bot->len > 1)) middle1 = bot;
	       if (bot != NULL) bot = bot -> next;
            }
#if 0
      }
#endif
    /* if the middle or the bottom lines match, we do not overwrite */

   if (bot != NULL) n2 = bot->len;
   if (middle1 == NULL) overwrite = 0; 
   else
     {
	n1 = middle1->len;
	for (r = r1; r <= r2; r++)
	  {
	     s = Screen[r - 1].line;
	     if (((middle1 == s) && (n1 > 1) && (n1 == s->len)
		  && (middle1->data[n1 / 2] == s->data[n1 / 2]))
		 || ((bot != NULL) && (bot == s) && (n2 > 4) && (n2 == s->len)
		     && (bot->data[n2/2] == s->data[n2 / 2])))
	       {
		  overwrite = 0;
		  break;
	       }
	  }
     }
   
   while(r1 <= r2)
     {
	 if (top == NULL) break;
	 match = 0;
	 for(r = r1; r <= r2; r++)
	   {
	      sline = Screen[r - 1].line;
	      if (sline == top)
		{
		   if (r != r1)
		     {
			scroll_region_set = 0;
			scroll_up(r - r1, r1, r2);
		     }

		   r1++;
		   match = 1;
		   if (top != NULL) top = top->next;
		   break;
		}
	   }

	 if (match) continue;

	 /* not found so insert it */
	 if ((top != NULL) && (!overwrite))
	   {
	      srs = update_insert_line(r1, top);
	      if (!scroll_region_set) scroll_region_set = srs;
	   }
	 if (top != NULL) top = top->next;
	 r1++;
      }

   /*     if (scroll_region_set) reset_scroll_region(); */
   close_scroll();
   if (overwrite) touch_window();
}

static void fix_char_width(void)
{
   int i;
   for (i = 0; i < 32; i++) 
     {
	char_width[i] = 2;
	char_width[i + 128] = 3;
     }
   
   for (i = 32; i < 127; i++) 
     {
	char_width[i] = 1;
	char_width[i + 128] = 2;
     }
   char_width[127] = 2;
   char_width[255] = 3;
   
   if (Display_Eight_Bit > 0)
     {
	for (i = (Display_Eight_Bit > 127 ? Display_Eight_Bit : 128);
	     i < 256; i++)
	  {
	     char_width[i] = 1;
	  }
     }
   dis8bit = Display_Eight_Bit;
   if ((Selective_Display = Buffer_Local.sd) < 0)
     {
	char_width[(unsigned char) '\r'] = 3;
     }
   if (JED_CSI > 0) char_width[(unsigned char) JED_CSI] = 0;
   j_csi = JED_CSI;
}


void point_column(int n)
{
   register unsigned char *p, *pmax;
   register int i;
   int tab, w;

   FIX_CHAR_WIDTH;
   if (IS_MINIBUFFER) n -= Mini_Info.prompt_len;

    p = CLine->data;

    pmax = p + (CLine->len - 1);
    if (*pmax != '\n') pmax++;

    tab = Buffer_Local.tab;
    i = 0;
    n--;   /* start at 0 */
    while(p < pmax)
      {
         if ((*p == '\t') && tab)
	   {
	      i = tab * (i / tab + 1);
	   }
	 else 
	   {
	      w = char_width[*p];
	      i +=  w;
	      if (w == 0) p++;
	   }
	 
	 if ((i > n) || ((Selective_Display < 0) && (*p == '\r'))) break;
	 p++;
      }
    Point = (int) (p - CLine->data);
}

/* given a position in a line, return apparant distance from bol
   expanding tabs, etc... up to pos */
int calculate_column()
{
   register int i;
   int tab, w;
   register unsigned char ch, *pos, *this_pos;

   FIX_CHAR_WIDTH;
   pos = CLine->data;
    this_pos = pos + Point;
    i = 1;
    tab = Buffer_Local.tab;
    while(ch = *pos, pos++ < this_pos)
      {
	 if ((ch == '\t') && tab)
	   {
	      i = tab * ((i - 1)/tab + 1) + 1;  /* tab column tabs */
	   }
	 else 
	   {
	      w = char_width[ch];
	      i += w;
	      if (w == 0) pos++;
	   }
	 
	 if ((ch == '\r') && (Selective_Display < 0)) break;
      }

    if (IS_MINIBUFFER) i += Mini_Info.prompt_len;

    Screen_Col = i;
    return (i);
}

void point_cursor()
{
    int r, row, c = 1;
    Line *this;

    r = 0;

   for (row = Window->top; row < Window->top + Window->rows; row++)
     {
	 this = Screen[row-1].line;
	 if (this == NULL) break;
	 if ((this == CLine) || (this == &Eob_Line))
	   {
	      r = row;
	      break;
	   }
      }
   
    if (Point >= CLine->len)
      {
	  Point = CLine->len - 1;

	  if (Point < 0) Point = 0;
	  else if (*(CLine->data + Point) != '\n') Point++;
      }

    if (r)
      {
          c = calculate_column();
          c = c - Window->column + 1;
          if (c < 1) c = 1; else if (c > Window->width) c = Window->width;
          goto_rc(r , c);
          Screen_Row = r;
          Screen_Col = c;
      }
    else
      msg_error("CLine not found in Window!");

    if (!Cursor_Motion) Goal_Column = c;
}

void make_status_line(Window_Type *w)
{
   int i;
   unsigned int line, maxline;
   char *s, pstr[50], mrk, flag, spot, star, ubit = '-';
   Buffer *buf;

   if (w->top == Screen_Height) return;   /* minibuffer ? */
   buf = CBuf;
   CBuf = w->buffer;
   line = w->mark.n - CBuf->nup;
   maxline = CBuf->max_linenum;
   
   if (!User_Prefers_Line_Numbers)
     {
	if (Screen[w->top - 1].line == CBuf->beg)
	  {
	     if (Screen[w->top - 2 + w->rows].line == NULL) strcpy(pstr," (All)");
	     else strcpy(pstr," (Top)");
	  }

	else if (Screen[w->top - 2 + w->rows].line == NULL) strcpy(pstr, " (Bot)");
	else
	  {
	     sprintf(pstr, " (%d%%)", 
		     (int) ((line * 100L) / (long) maxline));
	  } 
     }
   else sprintf(pstr, " (%d/%d)", line, maxline);

   s = Screen[w->rows + w->top - 1].new;
   mrk = flag = spot = star = '-';
   if (CBuf->marks != NULL) mrk = 'm';
   if (CBuf->flags & FILE_MODIFIED) flag = 'd';
   if (CBuf->spots != NULL) spot = 's';
   if (CBuf->flags & BUFFER_TRASHED) star = '*';
   if (CBuf->flags & READ_ONLY) star = '%';
   if (CBuf->flags & UNDO_ENABLED) ubit = '+';

   sprintf(s,"-%c%c%c%c%c%c : %s: %s =(%s", 
	   star, star, mrk, flag, spot, ubit, JED_VERSION, CBuf->name, CBuf->mode_str);
   

   if (CBuf->flags & OVERWRITE_MODE) strcat(s, " Ovwrt");
   strcat (s, ")=");
   strcat(s, pstr);

   if (Defining_Keyboard_Macro) strcat(s, " [Macro]");
   if (CBuf->narrow != NULL) strcat (s, " [Narrow]");
   
   i = strlen(s);
   MEMSET(s + i, '-', w->width - i);
   /* for (i = strlen(s); i < w->width;) s[i++] = '-', s[i++] = '-'; */
   s[w->width] = 0;
   CBuf = buf;
}

int update_status_line(Window_Type *w)
{
    int r;

    if (w->top == Screen_Height) return(0);   /* minibuffer ? */
    r = w->rows + w->top - 1;
    make_status_line(w);
    if (!strncmp(Screen[r].old, Screen[r].new, Window->width)) return(0);

    tt_reverse_video();
    goto_rc(r + 1, 1);
    smart_puts(Screen[r].new, Screen[r].old, r + 1);
    tt_normal_video();
    update_screen_txt(r);
    Screen[r].n = 80;
    return(1);
}

/* if force then do update otherwise return 1 if update or 0 if not */
int update_1(Line *top, int force)
{
   int i;
   Window_Type *w;
   int did_eob = 0;
   
   Suspend_Screen_Update = 0;

   if (Batch 
       || (!force 
	   && (Executing_Keyboard_Macro || (Repeat_Factor != NULL)
	       || input_pending(&Number_Zero) || (Read_This_Character != NULL)))
       || (CBuf != Window->buffer))
       
     {
	/* Window->trashed = 1; */
	/* This messes up blink match and I don not think we need it if all
	   functions return the correct value.
	      */
	 return(0);
     }
   Window->mark.line = CLine;
   Window->mark.point = Point;
   Window->mark.n = LineNum + CBuf->nup;
   CBuf->linenum = LineNum;
   CBuf->max_linenum = Max_LineNum;
   
   /* if cursor moves just left right, do not update status line */
   if (!force && !Window->trashed && (Cursor_Motion != 1) 
	/* if % wanted, assume user is like me and gets annoyed with 
	   screen updates */
       && User_Prefers_Line_Numbers)
     {
	update_status_line(Window);
	return(1);
     }

   if (!Window->trashed && Cursor_Motion) return(1);

    w = Window;
    do
      {
	 did_eob = 0;
	 if (top == NULL) top = find_top();
	 Window->beg.line = top;

	 /* scroll the screen to optimal location */
	 do_scroll(top);

	 for (i = Window->top - 1; i < Window->rows + Window->top - 1; i++)
	   {
	      /* the next line is really optional */
	      if (!force && (Exit_From_MiniBuffer ||    /* service this first */
			     input_pending(&Number_Zero))) break;
	       
	      if ((Screen[i].line != top) || (Screen[i].flags)
		  || (Want_Eob && !did_eob && (i != Screen_Height - 1)))
		{
		   if (((top == NULL) || (top->len == 0)) 
		       && (Want_Eob && !did_eob && !(CBuf->flags & READ_ONLY)))
		     {
			display_line(&Eob_Line, i + 1);
			
			/* Screen[i].line = top; */
			did_eob = 1;
		     }
		   else display_line(top, i + 1);
		}
	      if (top != NULL) top = top->next;
	   }
	 
	 if (!force && input_pending(&Number_Zero))
	   {
	      while(Window != w) other_window();
	      Window->trashed = 1;  /* since cursor not pointed */
	      return(0);
	   }
	 else update_status_line(Window);
	 Window->trashed = 0;
	 other_window();
	 top = NULL;
	 /* if (!Window->trashed) top = Window->beg.line; else  top = NULL; */
	 
      } while(Window != w);
   return(1);
}

static int Mini_Ghost = 0;

void update_minibuffer(void)
{
   Window_Type *w;

   if (Executing_Keyboard_Macro) return;
   if (MiniBuffer != NULL)
     {
	w = Window;
	while (!IS_MINIBUFFER) other_window();
	if (*Message_Buffer) (void) input_pending(&Number_One);
	display_line(CLine, Screen_Height);
	while (w != Window) other_window();
	Mini_Ghost = 1;
     }
   else if (Mini_Ghost && !*Error_Buffer && !*Message_Buffer)
     {
	/* if < 0, it is a result of flush message so let it pass this round */
	if (Mini_Ghost < 0) Mini_Ghost = 1;
	else
	  {
	     display_line(NULL, Screen_Height);
	     if (!Window->trashed) point_cursor();
	     Mini_Ghost = 0;
	  }
     }
   else Mini_Ghost = ((*Message_Buffer) || (*Error_Buffer));
}

void do_dialog(char *b)
{
   int len = 0, dout, row = Screen_Height - 1;

   if (Batch) return;
   FIX_CHAR_WIDTH;
   if (! *b)
     {
	if(!KeyBoard_Quit) return;
	b = "Quit!";
     }
   if (MiniBuffer)
     {
	len = Mini_Info.prompt_len;
	Mini_Info.prompt_len = 0;
     }

   dout = output((unsigned char *) b, strlen(b), Screen_Height, Screen_Width, 1);
   if (MiniBuffer) Mini_Info.prompt_len = len;
   
   if ((b == Error_Buffer) || (KeyBoard_Quit))
     {
	beep();
	flush_input();
	if (MiniBuffer != NULL) (void) input_pending(&Number_One);
     }
   
   if (!dout)
     {
	if (Screen[row].n)
	  {
	     goto_rc(Screen_Height,1);
	     tt_del_eol();
	  }
	blank_line(row);
	Screen[row].n = 0;
      }
   else update_screen_txt(row);
   Mini_Ghost = -dout;
}

void update(Line *line)
{
   if (Batch) return;
   FIX_CHAR_WIDTH;
   if (Lang_Error && !(*Error_Buffer || KeyBoard_Quit)) lang_doerror(NULL);
   if (*Error_Buffer || KeyBoard_Quit)
     {
	do_dialog(Error_Buffer);
	KeyBoard_Quit = 0;
	Lang_Error = 0;
	Mini_Ghost = 1;
	(void) update_1(line, 1);
	update_minibuffer();
     }
   else if (*Message_Buffer)
     {
	if (!update_1(line, 0))
	  {
	     /* *Message_Buffer = 0; */
	     return;
	  }
	Mini_Ghost = 1;
	do_dialog(Message_Buffer);
	update_minibuffer();
     }
   else
     {
	update_minibuffer();
	if (!update_1(line, 0)) return;
     }
   point_cursor();
   *Error_Buffer = *Message_Buffer = 0;
}

/* search for the CLine in the SCreen and flag it as changed */
/* n = 0 means line was changed, n = 1 means it was destroyed */
void register_change(int n)
{
   int i;
   Window_Type *w;
   register Line *cl = CLine;
   
    if (No_Screen_Update)
      {
	  No_Screen_Update = 0;
          if (((n == CINSERT) || (n == CDELETE)) && (Window->next == Window))
	    {
	       /* Since no screen update, we are probably safe to do: */
	       /* Screen[Screen_Row - 1].flags = 1; */
	       return;
	    }
	  w = Window->next;  /* skip this window */
      }
    else w = Window;

    do
      {
	  for (i = w->top - 1; i < w->top + w->rows - 1; i++)
	    {
		if (Screen[i].line == cl)
		  {
		      Screen[i].flags = 1;
		      if ((n == NLDELETE) || (n == LDELETE)) Screen[i].line = NULL;
		  }
	    }
	  w = w->next;
      }
    while(w != Window);
}

void reset_display()
{
   int i;
#ifdef msdos
   int ab = Attribute_Byte;
   Attribute_Byte = 0x7;
#endif

   if (Batch) return;
     
#ifndef msdos
   send_string_to_term("\033[?6l");   /* normal origin mode */
#endif
   goto_rc(Screen_Height,1);
   tt_erase_line();
     
   
   for (i = 0; i < Screen_Height; i++)
     {
	FREE(Screen[i].old);
	FREE(Screen[i].new);
     }
#ifdef msdos
   Attribute_Byte = ab;
#endif
}

void init_display(int g)
{
   int i;
   char *old, *new = NULL;

   if (Batch) return;
   if (g) get_term_dimensions(&Screen_Width, &Screen_Height);
#ifndef msdos
   send_string_to_term("\033[?6h");  /* relative origin mode */
   reset_scroll_region();
   end_insert();
#else
      init_video();
#endif
   for (i = 0; i < Screen_Height; i++)
     {
	/* these are bigger to handle overflow in output routine in case
	   special char occurs at end as well as allowing code as in blank_line*/
	if ((NULL == (old = (char *) MALLOC(Screen_Width + 3)))
	    || (NULL == (new = (char *) MALLOC(Screen_Width + 3))))
	  {
	     exit_error("init_display(): malloc error.");
	  }

	Screen[i].line = NULL;
	Screen[i].flags = 1;
	Screen[i].n = 0;
	Screen[i].old = old;
	Screen[i].new = new;
	MEMSET(old, ' ', Screen_Width);
     }
}



int redraw()
{
    int row, center;
    Window_Type *w;
   
   if (Batch) return(0);
    cls();
#ifndef msdos
   send_string_to_term("\033[?6h");   /* relative origin mode */
   reset_scroll_region();
   end_insert();
#endif
    for (row = 0; row < Screen_Height; row++)
      {
	 Screen[row].n = 0;
	 Screen[row].flags = 1;
	 blank_line(row);
      }
    w = Window;
    center = Window->trashed;
    do
      {
	 w->trashed = 1;
	 w = w->next;
      }
    while(w != Window);
    if (!center) update(Window->beg.line); else update((Line *) NULL);
    return(0);
}


void recenter(int *np)
{
   Line *l = CLine;
   int n = *np;
   
   if (CBuf != Window->buffer) return;
   if ((n <= 0) || (n > Window->rows)) n = Window->rows / 2;
   while (--n)
     { 
	l = l->prev;
	if (l == NULL)
	  {
	     l = CBuf->beg;
	     break;
	  }
     }
   
   Window->trashed = 1;
   update(l);
}


int window_line(void)
{
   Line *top = find_top();
   int n = 1;
   
   if (CBuf != Window->buffer) return(0);
   while ((top != NULL) && (top != CLine))
     {
	n++;
	top = top->next;
     }
   return(n);
}

   
void touch_window()
{
    int i;

    for (i = 0; i < Window->rows; i++)
      {
	 Screen[i + Window->top - 1].flags = 1;
      }

    Window->trashed = 1;
}

void touch_screen()
{
   Window_Type *w;
   
   Suspend_Screen_Update = 0;
   w = Window;
   do
     {
	touch_window();
	Window = Window->next;
     }
   while(w != Window);
}
