/*
 *  Copyright (c) 1992, 1993 John E. Davis  (davis@amy.tch.harvard.edu)
 *  All Rights Reserved.
 */
#include <stdio.h>
#include <string.h>
#include "buffer.h"
#include "search.h"
int Case_Sensitive = 0;

/* #define upcase(ch) UPPER_CASE(ch) */
#define upcase(ch) (Case_Sensitive ? ch : UPPER_CASE(ch))

static int ind[256];

unsigned char *forw_search_region
    (register unsigned char *beg,unsigned char *end, unsigned char *key, int key_len)
{
   unsigned char char1;
   unsigned char *pos;
   int j, str_len, db;
   register unsigned char ch;

   str_len = (int) (end - beg);
   if (str_len < key_len) return (NULL);

   char1 = key[key_len - 1];
   beg += (key_len - 1);

   while(1)
     {
	while (beg < end)
	  {
	     ch = *beg;
	     db = ind[(int) ch];
	     if ((db < key_len) && ((ch = upcase(ch)) == char1)) break;
	     beg += ind[(unsigned char) ch];
	  }
	if (beg >= end) return(NULL);
	
	pos = beg - (key_len - 1);
	for (j = 0; j < key_len; j++)
	  {
	     ch = upcase(pos[j]);
	     if (ch != (unsigned char) key[j]) break;
	  }
	
	if (j == key_len) return(pos);
	beg += 1;
     }
}

unsigned char *back_search_region(unsigned char *beg,unsigned char *end,
				  unsigned char *key, int key_len)
{
   unsigned char ch, char1;
    int j, str_len, ofs;

    str_len = (int) (end - beg);
    if (str_len < key_len) return (NULL);
    end -= (key_len - 1);

    char1 = key[0];

    while(1)
      {
	 while ((beg <= end) && (ch = *end, ch = upcase(ch), ch != char1))
	   {
	      ofs = ind[(unsigned char) ch];
#ifdef msdos
	      /* This is needed for msdos segment wrapping problems */
	      if (beg + ofs > end) return(NULL);
#endif
	      end -= ofs;
	   }
	 if (beg > end) return(NULL);
	 for (j = 1; j < key_len; j++)
	   {
	      ch = upcase(end[j]);
	      if (ch != key[j]) break;
	   }
	 if (j == key_len) return(end);
	 end--;
      }
}

static int upcase_search_word(char *str, char *work, int dir)
{
   int i, maxi;
   int max = strlen(str);
   char *w;
   
   
   if (dir > 0) 
     {
	w = work;
     }
   else
     {
	maxi = max - 1;
	str = str + maxi;
	w = work + maxi;
     }
   
   for (i = 0; i < 256; i++) ind[i] = max;
   
   i = 0;
   while (i++ < max)
     {
	maxi = max - i;
	if (Case_Sensitive) 
	  {
	     *w = *str;
	     ind[(unsigned char) *str] = maxi;
	  }
	else
	  {
	     *w = UPPER_CASE(*str);
	     ind[(unsigned char) *w] = maxi;
	     ind[(unsigned char) LOWER_CASE(*str)] = maxi;
	  }
	str += dir; w += dir;
     }
   work[max] = 0;
   return(max);
}

int search(char *str, int dir, int n)
{
   unsigned char *beg, *end, *p, work[132];
   Line *line;
   int key_len;
   unsigned int num = 0;

   key_len = upcase_search_word(str, (char *) work, dir);
   if (!key_len) return(0);
   line = CLine;

   if (dir == 1)
     {
	beg = line->data + Point;
	end = line->data + line->len;
	do
	  {
	     if (NULL !=
		 (p = forw_search_region(beg, end, (unsigned char *) work,
					 key_len)))
	       {
		  CLine = line;
		  LineNum += num;
		  Point = (int) (p - line->data);
		  return(1);
	       }
	     line = line->next; num++;
	     if (line == NULL) return(0);
	     beg = line->data;
	     end = line->data + line->len;
	  }
	while(--n);
     }
   else if (dir == -1)
     {
	beg = line->data - 1;
	if (Point) end = line->data + Point - 1; else end = beg;
	do
	  {
	     if (NULL !=
		 (p = back_search_region(beg, end, (unsigned char *) work,
					 key_len)))
	       {
		  CLine = line;
		  LineNum -= num;

		  Point = (int) (p - line->data);
		  return(1);
	       }
	     line = line->prev;
	     num++;
	     if (line == NULL) return(0);
	     beg = line->data;
	     end = line->data + line->len;
	  }
	while (--n);
     }
   return(0);
}

int search_forward(char *s)
{
   return( search(s, 1, 0));
}
int search_backward(char *s)
{
   return( search(s, -1, 0));
}

int forward_search_line(char *s)
{
   return( search(s, 1, 1));
}

int backward_search_line(char *s)
{
   return( search(s, -1, 1));
}

int bol_search(char *str, int dir)
{
   Line *this;
   int max, n;
   unsigned int num;
   unsigned char *s, ch1, ch2;
   
   max = strlen(str);
   if (dir > 0)
     {
	num = 1;
	this = CLine->next;
     }
   else
     {
	if (Point == 0) 
	  {
	     this = CLine->prev;
	     num = 1;
	  }
	else
	  {
	     this = CLine;
	     num = 0;
	  }
     }
   
	
   while (this != NULL)
     {
	if (max <= this->len)
	  {
	     s = this->data;
	     for (n = 0; n < max; n++)
	       {
		  ch1 = upcase(s[n]);
		  ch2 = upcase(str[n]);
		  if (ch1 != ch2) break;
	       }
	     if (n == max)
	       {
		  Point = 0;
		  CLine = this;
		  if (dir > 0) LineNum += num; else LineNum -= num;
		  return(1);
	       }
	  }
	num++;
	if (dir > 0) this = this->next; else this = this->prev;
     }
   return(0);
}

int bol_fsearch(char *s)
{
   return bol_search(s, 1);
}

int bol_bsearch(char *s)
{
   return bol_search(s, -1);
}
