/*
 *  Copyright (c) 1992, 1993 John E. Davis  (davis@amy.tch.harvard.edu)
 *  All Rights Reserved.
 */
#include <stdio.h>

#ifdef sequent
# define strchr index
  extern char *my_strstr(char *, char *);
# define strstr(a,b) my_strstr(a, b);
#else
# include <stdlib.h>
#endif

#include <string.h>

#ifdef ultrix
extern int system(char *);       /* used for slang function 'system' */
#else
extern int system(const char *);       /* used for slang function 'system' */
#endif


#include "slang.h"
#include "intrin.h"
#include "malloc.h"

#ifdef msdos
#  define SLANG_SYSTEM_NAME "_IBMPC"
#else
#  ifdef VMS
#    define SLANG_SYSTEM_NAME "_VMS"
#  else
#    define SLANG_SYSTEM_NAME "_UNIX"
#  endif
#endif

/* If non null, these call C functions before and after a slang function. */
void (*Lang_Enter_Function)(char *) = NULL;
void (*Lang_Exit_Function)(char *) = NULL;


#ifdef SLANG_STATS
#define MAKE_INTRINSIC(n, f, out, in)        \
    {0, n, {((out | (in << 4)) << 8) | LANG_INTRINSIC, (long) f}}
       
#define MAKE_VARIABLE(n, v, t, r)     \
    {0, n, {(LANG_IVARIABLE + r) | (t << 8), (long) v}}

#else
#define MAKE_INTRINSIC(n, f, out, in)        \
    {n, {((out | (in << 4)) << 8) | LANG_INTRINSIC, (long) f}}
       
#define MAKE_VARIABLE(n, v, t, r)     \
    {n, {(LANG_IVARIABLE + r) | (t << 8), (long) v}}
#endif

#define LANG_MAX_NAME_LEN 30
/* maximum length of an identifier */
/* first char in identifiers is the hash */

typedef struct Lang_Name_Type
  {
#ifdef SLANG_STATS
     int n;			       /* number of times referenced */
#endif
     char name[LANG_MAX_NAME_LEN + 2]; /* [0] is hash */
     Lang_Object_Type obj;
  }
Lang_Name_Type;

   
extern Lang_Name_Type Lang_Intrinsic_Name_Table[];

#define LANG_MAX_SYMBOLS 500
/* maximum number of global symbols--- slang builtin, functions, global vars */

Lang_Name_Type Lang_Name_Table[LANG_MAX_SYMBOLS];
Lang_Name_Type *Lang_Local_Variable_Table;
int Local_Variable_Number;
#define MAX_LOCAL_VARIABLES 30

int Lang_Break_Condition = 0;	       /* true if any one below is true */
int Lang_Break = 0;
int Lang_Return = 0;
int Lang_Continue = 0;

/* this stack is used by the inner interpreter to execute top level
   interpreter commands which by definition are immediate so stack is
   only of maximum 4: the 2 blocks, else nnd terminator */
Lang_Object_Type Lang_Interp_Stack_Static[4];
Lang_Object_Type *Lang_Interp_Stack_Ptr = Lang_Interp_Stack_Static;
Lang_Object_Type *Lang_Interp_Stack = Lang_Interp_Stack_Static;

/* maximum size of run time stack */
#define LANG_MAX_STACK_LEN 1000
Lang_Object_Type Lang_Run_Stack[LANG_MAX_STACK_LEN];
int Lang_Run_Stack_Depth = -1;

/* Might want to increase this. */
#define MAX_LOCAL_STACK 200
Lang_Object_Type Local_Variable_Stack[MAX_LOCAL_STACK];

Lang_Object_Type *Local_Variable_Frame = Local_Variable_Stack;

volatile int Lang_Error = 0;
int Lang_Traceback = 0;		       /* non zero means do traceback */

extern int inner_interp(Lang_Object_Type *addr);

int Lang_Defining_Function = 0;   /* true if defining a function */
Lang_Object_Type *Lang_Function_Body;
Lang_Object_Type *Lang_FBody_Ptr;
int Lang_FBody_Size = 0;

#define LANG_MAX_BLOCKS 30
/* max number of nested blocks--- was 10 but I once exceeded it! */

typedef struct Lang_Block_Type
  {
     int size;                         /* current nuber of objects malloced */
     Lang_Object_Type *body;           /* beginning of body definition */
     Lang_Object_Type *ptr;            /* current location */
  }
Lang_Block_Type;

int Lang_Defining_Block = 0;   /* true if defining a block */
Lang_Block_Type Lang_Block_Stack[LANG_MAX_BLOCKS];
Lang_Object_Type *Lang_Block_Body;
int Lang_BBody_Size;

int Lang_Block_Depth = -1;

Lang_Object_Type *Lang_Object_Ptr = Lang_Interp_Stack_Static;
/* next location for compiled obj -- points to interpreter stack initially */

/* useful macro to tell if string should be freed after its use. */
#define IS_DATA_STRING(obj)\
   ((((obj).type & 0xFF) == LANG_DATA) && (((obj).type >> 8) == STRING_TYPE))

#ifdef VMS
int (*Lang_Error_Routine)(char *) = NULL;
#else
int (*Lang_Error_Routine)(char *) = (int (*)(char *)) NULL;
#endif

void lang_doerror(char *error)
{
   char err[80]; char *str;

   if (!Lang_Error) Lang_Error = UNKNOWN_ERROR;
   *err = 0;
   switch(Lang_Error)
     {
	case (UNDEFINED_NAME): str = "Undefined_Name"; break;
	case (SYNTAX_ERROR): str = "Syntax_Error"; break;
	case (STACK_OVERFLOW): str = "Stack_Overflow"; break;
	case (STACK_UNDERFLOW): str = "Stack_Underflow"; break;
	case (DUPLICATE_DEFINITION): str = "Duplicate_Definition"; break;
	case (TYPE_MISMATCH): str = "Type_Mismatch"; break;
	case(READONLY_ERROR): str = "Variable is read only."; break;
	case (MALLOC_ERROR) : str = "S-Lang: Malloc Error."; break;
      case USER_BREAK: strcpy(err, "User Break!"); break;
	case (INTRINSIC_ERROR): str = "Intrinsic Error"; break;
	/* application code should handle this */
	default: if (error != NULL) str = error; else str = "Unknown Error.";
     }

   if (*err == 0) sprintf(err, "S-Lang Error: %s", str);
   
   if (Lang_Error_Routine == NULL)
     fputs(err, stderr);
   else (*Lang_Error_Routine)(err);
}


#ifdef SLANG_LOAD_FILE
void (*Lang_Message_Routine)(char *) = (void (*)()) puts;
char *SLang_Library = NULL;

int lang_load_file(char *file)
{
   char filebuf[255];
   int count = 0;
   FILE *fp;
   char *err;
#ifdef msdos
   char *slash = "\\";
#else 
#ifdef VMS
   char *slash = "";
#else
   char *slash = "/";
#endif
#endif
   
   strcpy(filebuf, "Loading "); strcat(filebuf, file);
   (*Lang_Message_Routine)(filebuf);

   if (NULL == (fp = fopen(file,"r")))
     {
	if (NULL != SLang_Library)
	  {
	     strcpy(filebuf, SLang_Library);
	     strcat(filebuf, slash);
	     strcat(filebuf, file);
	     fp = fopen(filebuf, "r");
	  }
	if (fp == NULL)	
	  {
	     lang_doerror(filebuf);
	     return(0);
	  }
     }

   while(NULL != fgets(filebuf, 254, fp))
     {
	count++;
	err = interpret(filebuf);
	if (Lang_Error)
	  {
	     sprintf(filebuf, "Error: line %d: %s", count, err);
	     (*Lang_Message_Routine)(filebuf);
	     break;
	  }
     }

   fclose(fp);

   return( !Lang_Error );
}
#endif

int lang_pop(Lang_Object_Type *x)
{
   register int depth = Lang_Run_Stack_Depth--;
   register Lang_Object_Type *y;

    if (depth < 0)
      {
	 x->type = 0;
	 Lang_Error = STACK_UNDERFLOW;
	 Lang_Run_Stack_Depth = -1;
	 return 0;
      }

   y = Lang_Run_Stack + depth;
   x->type = y->type;
   x->value = y->value;
   return(1);

}

void lang_push(Lang_Object_Type *x)
{
   register int depth;
   register Lang_Object_Type *y;

   if (Lang_Error) return;
   depth = ++Lang_Run_Stack_Depth;

    /* flag it now */
   if (Lang_Run_Stack_Depth == LANG_MAX_STACK_LEN)
      Lang_Error = STACK_OVERFLOW;

   y = Lang_Run_Stack + depth;
   y->type = x->type;
   y->value = x->value;
}

void lang_free_branch(Lang_Object_Type *p)
{
   short type; /* s_type = LANG_LITERAL | (STRING_TYPE << 8); */
   
   while(1)
     {
        type = (p->type);
	if ((type & 0xFF) == LANG_BLOCK)
	  {
	     lang_free_branch((Lang_Object_Type *) p->value);
	     FREE(p->value);
	  }
	/* else if (type == s_type) FREE(p->value);
	 This fails because objects may be attached to these strings */
	else if (type == 0) break;
	p++;
     }
}

int lang_pop_integer(int *i)
{
   Lang_Object_Type obj;

   if (!lang_pop(&obj) || ((obj.type >> 8) != INT_TYPE))
     {
	if (!Lang_Error)
	  {
	     lang_push(&obj);
	     Lang_Error = TYPE_MISMATCH;
	  }
	return(0);
     }

   *i = (int) obj.value;
   return(1);
}

#ifdef FLOAT_TYPE
int lang_pop_float(float *x)
{
   Lang_Object_Type obj;
   unsigned char stype;

   if (!lang_pop(&obj)) return(0);
   stype = obj.type >> 8;

   if (stype == INT_TYPE) *x = (float) obj.value;
   else if (stype == FLOAT_TYPE) *x = *(float *) &obj.value;
   else
     {
	lang_push(&obj);
	Lang_Error = TYPE_MISMATCH;
	return(0);
     }
   return(1);
}

void lang_push_float(float x)
{
   Lang_Object_Type obj;

   obj.type = LANG_DATA | (FLOAT_TYPE << 8);
   obj.value = *(long *) &x;
   lang_push (&obj);
}

#endif

int lang_pop_string(char **s, int *data)
{
   Lang_Object_Type obj;

   if (!lang_pop(&obj) || ((obj.type >> 8) != STRING_TYPE))
     {
	lang_push(&obj);
	if (!Lang_Error) Lang_Error = TYPE_MISMATCH;
	return(0);
     }

   *s = (char *) obj.value;
   /* return whether or not this should be freed after its use. */
   if ((obj.type & 0xFF) == LANG_DATA) *data = 1; else *data = 0;
   return(1);
}

void lang_push_integer(int i)
{
   Lang_Object_Type obj;

   obj.type = LANG_DATA | (INT_TYPE << 8);
   obj.value = (long) i;
   lang_push (&obj);
}

long lang_make_string(char *str)
{
   char *ptr;

   if (NULL == (ptr = (char *) MALLOC(strlen(str) + 1)))
     {
	Lang_Error = MALLOC_ERROR;
	/* lang_doerror("malloc error in lang_make_string."); */
	return(NULL);
     }
   strcpy((char *) ptr, str);
   return((long) ptr);
}

void lang_push_string(char *t)
{
   Lang_Object_Type obj;

   if (NULL == (obj.value = (long) lang_make_string((char *) t))) return;
   obj.type = LANG_DATA | (STRING_TYPE << 8);
   lang_push(&obj);
}

/* array types */
typedef struct Array_Type
{
   int dim;			       /* # of dims (max 3) */
   int x,y,z;			       /* actual dims */
   long ptr;			       /* address of buffer */
   unsigned char type;		       /* int, float, etc... */
} Array_Type;



/* if ptr == NULL then malloc space.  Otherwise assume space is at ptr */
Array_Type *lang_create_array(long *ptr, int dim, int d0, int d1, int d2, char *t)
{
   unsigned long n, size;
   unsigned char type;
   Array_Type *at;

   switch (*t)
     {
	case 'i': type = INT_TYPE; size = sizeof(int); break;
	case 's': type = STRING_TYPE; size = sizeof(char *); break;
#ifdef FLOAT_TYPE
	case 'f': type = FLOAT_TYPE; size = sizeof(float); break;
#endif
	default: return (NULL);
     }

   n = d0;
   n = n * d1;
   n = n * d2;

   if (NULL == (at = (Array_Type *) MALLOC(sizeof(Array_Type))))
     {
	return NULL;
     }

   if ((*ptr == NULL) && ((long) NULL == (*ptr = (long) CALLOC(n, size))))
     {
	return NULL;
     }
   at->ptr = *ptr;
   at->dim = dim;
   at->x = d0; at->y = d1; at->z = d2;
   at->type = type;
   return(at);
}

void create_array(void)
{
   int dim, d0, d1, d2, tstr_data;
   char *tstr;
   long ptr;
   Lang_Object_Type obj;
   Array_Type *at;

   if (!lang_pop_integer(&dim)) return;

   if (dim > 3)
     {
	lang_doerror("Array size not supported.");
	return;
     }

   d1 = d0 = d2 = 1;
   switch(dim)
     {
      case 3: lang_pop_integer(&d2);
      case 2: lang_pop_integer(&d1);
      case 1: lang_pop_integer(&d0);
     }

   if (!lang_pop_string(&tstr, &tstr_data)) return;

   ptr = (long) NULL;
   at = lang_create_array((long *) &ptr, dim, d0, d1, d2, tstr);
   if (tstr_data) FREE(tstr);

   if (at == NULL)
     {
	lang_doerror("Unable to create array.");
	return;
     }

   obj.type = LANG_DATA | (ARRAY_TYPE << 8);
   obj.value = (long) at;
   lang_push(&obj);
}

void free_array(void)
{
   Lang_Object_Type obj;

   if (!lang_pop(&obj)) return;

   if ((obj.type >> 8) != ARRAY_TYPE)
     {
	Lang_Error = TYPE_MISMATCH;
	return;
     }

   FREE( ((Array_Type *) obj.value)->ptr );
   FREE(obj.value);
}

Array_Type *pop_array(void)
{
   Lang_Object_Type obj;

   if (!lang_pop(&obj)) return(NULL);

   if ((obj.type >> 8) != ARRAY_TYPE)
     {
	Lang_Error = TYPE_MISMATCH;
	return(NULL);
     }

   return (Array_Type *) obj.value;
}

unsigned int compute_array_offset(Array_Type *at)
{
   int elem[3], el, x[3], d, dim;
   unsigned int off;

   if (Lang_Error) return(0);
   dim = at->dim;
   x[0] = at->x; x[1] = at->y; x[2] = at->z;
   elem[0] = elem[1] = elem[2] = 0;
   d = dim;

   while (d--)
     {
	if (!lang_pop_integer(&el)) return(0);
	if ((el > x[d]) || (el < 1))
	  {
	     lang_doerror("Array dims out of bounds.");
	     return(-1);
	  }
	elem[d] = el - 1;
     }

   off = 0;
   d = 0;
#if 0
   while (d < dim)
     {
	off += off * x[d] + elem[d];
	d++;
     }
#endif
   off = (elem[0] * x[1] + elem[1]) * x[2] + elem[2];
   
   return(off);
}


void array_putelem()
{
   Array_Type *at;
   unsigned int off;
   int sdat, i, *ip;
   char *str, *newstr, **sp;
   unsigned char *p;
#ifdef FLOAT_TYPE
   float f, *fp;
#endif

   if (NULL == (at = pop_array())) return;

   off = compute_array_offset(at);
   if (Lang_Error) return;

   p = (unsigned char *) at->ptr;
   switch(at->type)
     {
	case INT_TYPE:
	  if (!lang_pop_integer(&i)) return;
	  ip = (int *) (off + (int *) p);
	  *ip = i; break;

	case STRING_TYPE:
	  if (!lang_pop_string(&str, &sdat)) return;
	  newstr = (char *) lang_make_string(str);
	  if (sdat) FREE(str);
	  sp = (char **)(off + (char **) p);

	  if (NULL != *sp) FREE(*sp);
	  *sp = newstr;
	  break;

#ifdef FLOAT_TYPE
	case FLOAT_TYPE:
	  if (!lang_pop_float(&f)) return;
	  fp = off + (float *) p;
	  *fp = f;
	  break;
#endif
	default: lang_doerror("Corrupted Array.");
     }
   return;
}

void array_push_element(Array_Type *at, int off)
{
   unsigned char *p;
   p = (unsigned char *) at->ptr;
   switch(at->type)
     {
	case INT_TYPE: lang_push_integer((int) *(((int *) p) + off)); break;
	case STRING_TYPE: lang_push_string((char *) *(((char **)p) + off)); break;
#ifdef FLOAT_TYPE
	case FLOAT_TYPE: lang_push_float((float) *(((float *)p) + off)); break;
#endif
	default: lang_doerror("Internal Error in array element.");
     }
}

void array_getelem()
{
   Array_Type *at;
   unsigned int off;

   if (NULL == (at = pop_array())) return;

   off = compute_array_offset(at);
   if (Lang_Error) return;
   array_push_element(at, off);
}

/* builtin stack manipulation functions */

void do_pop(void)
{
   Lang_Object_Type x;
   if (!lang_pop(&x)) return;

   if (IS_DATA_STRING(x)) FREE(x.value);
}

int do_dup(void)
{
   Lang_Object_Type x;
   if (!lang_pop(&x)) return(0);
   lang_push(&x);
   if ((x.type >> 8) == STRING_TYPE)
     lang_push_string((char *) x.value); else lang_push (&x);
   return(1);
}

void do_strcat(void)
{
   char *a, *b, *c;
   int len;
   Lang_Object_Type obj;
   int adata, bdata;

   if (!lang_pop_string(&b, &bdata) || !lang_pop_string(&a, &adata)) return;

   len = strlen(a) + strlen(b);
   if (NULL != (c = (char *) MALLOC(len + 1)))
     {
	strcpy(c, a);
	strcat(c, b);
	
	/* instead of going throug push string, push it directly */
	obj.type = LANG_DATA | (STRING_TYPE << 8);
	obj.value = (long) c;
	lang_push(&obj);
     }
   else Lang_Error = MALLOC_ERROR; /* lang_doerror("Lang Malloc error."); */
   
   if (adata) FREE(a);
   if (bdata) FREE(b);
}

void do_strtrim(void)
{
   char *a, *beg, *end, *c, *new, ch;
   int len;
   Lang_Object_Type obj;
   int adata;

   if (!lang_pop_string(&a, &adata)) return;

   len = strlen(a);
   beg = a;
   end = a + (len - 1);
   while (ch = *beg, (ch == ' ') || (ch == '\t') || (ch == '\n')) beg++;
   while (end >= beg)
     {
	ch = *end;
	if ((ch == ' ') || (ch == '\t') || (ch == '\n')) end--;
	else break;
     }
   end++;
   len = (int) (end - beg);
   /* instead of going throug push string, push it directly */
   if (NULL != (c = (char *) MALLOC(len + 1)))
     {
	new = c;
	while (beg < end) *c++ = *beg++;
	*c = 0;

	obj.type = LANG_DATA | (STRING_TYPE << 8);
	obj.value = (long) new;
	lang_push(&obj);
     }
   else Lang_Error = MALLOC_ERROR; /* lang_doerror("Lang Malloc error."); */
   
   if (adata) FREE(a);
}

/* returns the position of substrin in a string or null */
void do_issubstr(void)
{
   char *a, *b, *c;
   int adata, bdata, n;

   if (!lang_pop_string(&b, &bdata) || !lang_pop_string(&a, &adata)) return;

   if (NULL == (c = strstr(a, b))) n = 0; else n = 1 + (int) (c - a);

   if (adata) FREE(a);
   if (bdata) FREE(b);
   lang_push_integer (n);
}

/* returns to stack string at pos n to n + m of a */
void do_substr(void)
{
   char *a;
   int adata, n, m;
   char b[255];

   if (!lang_pop_integer(&m) || !lang_pop_integer(&n) || (!lang_pop_string(&a, &adata))) return;

   *b = 0;
   if (m > 0)
     {
	strncpy(b, a + (n - 1), 254);
	if (m > 255) m = 255;
	b[m] = 0;
     }
   if (adata) FREE(a);
   lang_push_string(b);
}
/* substitute char m at positin string n in string*/
void do_strsub(void)
{
   char *a;
   int adata, n, m;
   char b[255];

   if (!lang_pop_integer(&m) || !lang_pop_integer(&n) || (!lang_pop_string(&a, &adata))) return;

   strncpy(b, a, 254);
   b[254] = 0;
   if (adata) FREE(a);
   if ((n < 1) || (n > 254)) n = 254;
   b[n-1] = (char) m;
   lang_push_string(b);
}

void do_strup(void)
{
   char *a;
   int adata;

   if (!lang_pop_string(&a, &adata)) return;
   lang_push_string(a);
   if (adata) FREE(a);
   
   a = (char *) Lang_Run_Stack[Lang_Run_Stack_Depth].value;
   while (*a)
     {
	if ((*a >= 'a') && (*a <= 'z')) *a -= 32;
	a++;
     }
}

void do_strlow(void)
{
   char *a;
   int adata;

   if (!lang_pop_string(&a, &adata)) return;
   lang_push_string(a);
   if (adata) FREE(a);
   
   a = (char *) Lang_Run_Stack[Lang_Run_Stack_Depth].value;
   while (*a)
     {
	if ((*a >= 'A') && (*a <= 'Z')) *a += 32;
	a++;
     }
}

void do_strcmp(void)
{
   char *a, *b;
   int adata, bdata, i;

   if (!lang_pop_string(&b, &bdata) || !lang_pop_string(&a, &adata)) return;

   i = strcmp(a, b);

   if (adata) FREE(a);
   if (bdata) FREE(b);
   lang_push_integer (i);
}

void do_strncmp(void)
{
   char *a, *b;
   int adata, bdata, i, n;
   
   if (!lang_pop_integer(&n) || !lang_pop_string(&b, &bdata) || !lang_pop_string(&a, &adata)) return;

   i = strncmp(a, b, n);

   if (adata) FREE(a);
   if (bdata) FREE(b);
   lang_push_integer (i);
}

void do_strlen(void)
{
   char *a;
   int adata, i;

   if (!lang_pop_string(&a, &adata)) return;

   i = strlen(a);

   if (adata) FREE(a);
   lang_push_integer (i);
}

int do_isdigit(char *what)
{
   if ((*what >= '0') && (*what <= '9')) return(1); else return(0);
}

/* convert object to integer form */
void do_int(void)
{
   Lang_Object_Type x;
   int i;
   unsigned char stype;

   if (!lang_pop(&x)) return;
   stype = x.type >> 8;

   if (stype == INT_TYPE)
     {
	lang_push(&x);
	return;
     }

   else if (stype == STRING_TYPE)
     {
	i = (int) *(unsigned char *) x.value;
	if (IS_DATA_STRING(x)) FREE(x.value);
     }
#ifdef FLOAT_TYPE
   else if (stype == FLOAT_TYPE)
     {
	i = (int) *(float *) &x.value;
     }
#endif
   else
     {
	Lang_Error = TYPE_MISMATCH;
	return;
     }
   lang_push_integer(i);
}

/* Conver string to integer */
void do_integer(void)
{
   char *a;
   int adata, i;

   if (!lang_pop_string(&a, &adata)) return;

   /* Should check for parse error here but later */
   i = atoi(a);

   if (adata) FREE(a);
   lang_push_integer (i);
}

#ifdef FLOAT_TYPE
void do_float(void)
{
   char *a;
   int adata;
   float f;

   if (!lang_pop_string(&a, &adata)) return;

   /* Should check for parse error here but later. */
   f = atof(a);

   if (adata) FREE(a);
   lang_push_float (f);
}
#endif
/* convert integer to a sring of length 1 */
void do_char(void)
{
   char ch, buf[2];
   int x;

   if (!lang_pop_integer(&x)) return;

   ch = (char) x;
   buf[0] = ch;
   buf[1] = 0;
   lang_push_string((char *) buf);
}

/* format object into a string */
void do_string(void)
{
   Lang_Object_Type x;
   char buf[255];

   if (!lang_pop(&x)) return;

   if ((x.type >> 8) == STRING_TYPE)
     {
	lang_push(&x);
	return;
     }
#ifndef FLOAT_TYPE
   sprintf(buf, "%d", (int) x.value);
#else
   sprintf(buf, "%.6g",  (float) *(float *) &x.value);
#endif
   lang_push_string((char *) buf);
}

static void call_funptr(Lang_Name_Type *);

void lang_push_variable(Lang_Object_Type *obj)
{
   unsigned char subtype;
   Lang_Object_Type new;
   long value;

   subtype = obj->type >> 8;
   value = obj->value;

   if (subtype == STRING_TYPE)
     {
	lang_push_string((char *) value);
	return;
     }
   else if (subtype == LANG_OBJ_TYPE)
     {
	call_funptr((Lang_Name_Type *) value);
	return;
     }
   
   new.type = obj->type;
   new.value = value;
   lang_push(&new);
}

/* local and global variable assignments */

/* value contains either the offset of data for local variables or
   location of object_type for global ones.  For strings, we have to be
   careful.  Literal (constant) strings which are already attached to these
   variables are not to be freed--- only those of type data (dynamic).
   There is no need to create new strings since they come from the stack.

   Note that strings appear on the stack in two forms:  literal and dynamic.
   Literal strings are constants.  Dynamic ones are created by, say, dup, etc.
   They are freed only by routines which eat them.  These routines must
   check to see if they are not literal types before freeing them.  The only
   other way they are freed is when they are on the local variable stack, e.g.,
   (assigned to local variables) and the function exits freeing them.

   Define Macro to do this:  (defined above)

#define IS_DATA_STRING(obj)\
   ((((obj).type & 0xFF) == LANG_DATA) && (((obj).type >> 8) == STRING_TYPE))
*/


/* pop a data item from the stack and return a pointer to it.
   Strings are not freed from stack so use another routine to do it.

   In addition, I need to make this work with the array types.  */

long *pop_pointer(unsigned short *type)
{
   Lang_Object_Type obj;
   long *val;
   int depth;

   if (!lang_pop(&obj)) return(0);

   /* use this because the stack is static but obj is not.
      do not even try to make it static either. See the intrinsic
      routine for details */
   depth = Lang_Run_Stack_Depth + 1;
   val = (long *) &Lang_Run_Stack[depth].value;
   *type = Lang_Run_Stack[depth].type;

   /* The assumption is that only data and literals are on stack */
   if ((*type >> 8) == STRING_TYPE) return((long *) *val);
   return((long *) val);
}

void lang_do_eqs(Lang_Object_Type *obj)
{
   int y;
   register unsigned char type;
   register Lang_Object_Type *addr;
   register long val;
   

   type = obj->type >> 8;
   val = (long) obj->value;   
   
   /* calculate address */
   switch(type)
     {
      case LANG_LEQS: 
      case LANG_LPP: 
      case LANG_LMM: 
      case LANG_LPEQS: 
      case LANG_LMEQS:
	addr = Local_Variable_Frame - (int) val;
	break;
      case LANG_GEQS: 
      case LANG_GPP: 
      case LANG_GMM: 
      case LANG_GPEQS: 
      case LANG_GMEQS: 
	val = ((Lang_Name_Type *) val)->obj.value;
	addr = (Lang_Object_Type *) val; 
	break;
      default: 
	addr = &((Lang_Name_Type *) val)->obj;
	val = addr->value;
     }

   switch (type)
     {
      case LANG_LEQS: 
      case LANG_GEQS: 
	if (IS_DATA_STRING(*addr)) FREE(addr->value);
	lang_pop(addr);
	return;
     }
      
   /* everything else applies to integers */
   if (INT_TYPE != (addr->type >> 8)) 
     {
	Lang_Error = TYPE_MISMATCH;
	return;
     }
   
   y = 1;
   switch (type)
     {
      case LANG_LPEQS: 
      case LANG_GPEQS: 
	if (!lang_pop_integer(&y)) return;
	/* drop */
      case LANG_LPP: 
      case LANG_GPP: 
	addr->value += y;
	break;
	
      case LANG_GMEQS: 
      case LANG_LMEQS: 
	if (!lang_pop_integer(&y)) return;
	/* drop */
      case LANG_LMM: 
      case LANG_GMM: 
	addr->value -= y;
	break;
	
      case LANG_IEQS: 
	if (!lang_pop_integer(&y)) return;
	*(int *) val = y;
	break;
	
      case LANG_IPEQS: 
	if (!lang_pop_integer(&y)) return;
	/* drop */
      case LANG_IPP: 
	*(int *) val += y;
	break;
	
      case LANG_IMEQS:
	if (!lang_pop_integer(&y)) return;
	/* drop */
      case LANG_IMM: 
	*(int *) val -= y;
	break;
      default: 
	Lang_Error = UNKNOWN_ERROR;
     }
}

/* intrinsic functions are passed parameters by reference. The function
   pointer is given by obj->value. */
void lang_do_intrinsic(Lang_Object_Type *objf)
{
   long (*f)();
   void (*fv)();
   long ret;
   long *p1, *p2, *p3, *p4, *p5;
   unsigned short p1type, p2type, p3type, p4type, p5type;
   unsigned char type;
   int argc;
#ifdef FLOAT_TYPE
   float xf;
   float (*ff)();
   ff = (float (*)()) objf->value;
#endif

   f = (long (*)()) objf->value;
   fv = (void (*)()) objf->value;

   p5type = p4type = p3type = p2type = p1type = 0;
   argc = LANG_INTRINSIC_ARGC(*objf);
   type = LANG_INTRINSIC_TYPE(*objf);

   switch (argc)
     {
	case 5: p5 = pop_pointer(&p5type);
	case 4: p4 = pop_pointer(&p4type);
	case 3: p3 = pop_pointer(&p3type);
	case 2: p2 = pop_pointer(&p2type);
	case 1: p1 = pop_pointer(&p1type);
     }
   
   /* I need to put a setjmp here so to catch any long jmps that occur
      in the user program */
   if (!Lang_Error) switch (argc)
     {

	case 0:
	  if (type == VOID_TYPE) (void) (*fv)();
#ifdef FLOAT_TYPE
	  else if (type == FLOAT_TYPE) xf = (*ff)();
#endif
	  else ret = (*f)();
	  break;

	case 1:
	  if (type == VOID_TYPE) (void) (*fv)(p1);
#ifdef FLOAT_TYPE
	  else if (type == FLOAT_TYPE) xf = (*ff)(p1);
#endif
	  else ret = (*f)(p1);
	  break;

	case 2:
	  if (type == VOID_TYPE) (void) (*fv)(p1, p2);
#ifdef FLOAT_TYPE
	  else if (type == FLOAT_TYPE) xf = (*ff)(p1, p2);
#endif
	  else ret = (*f)(p1, p2);
	  break;

	case 3:
	  if (type == VOID_TYPE) (void) (*fv)(p1, p2, p3);
#ifdef FLOAT_TYPE
	  else if (type == FLOAT_TYPE) xf = (*ff)(p1, p2, p3);
#endif
	  else ret = (*f)(p1, p2, p3);
	  break;

	case 4:
	  if (type == VOID_TYPE) (void) (*fv)(p1, p2, p3, p4);
#ifdef FLOAT_TYPE
	  else if (type == FLOAT_TYPE) xf = (*ff)(p1, p2, p3, p4);
#endif
	  else ret = (*f)(p1, p2, p3, p4);
	  break;

	case 5:
	  if (type == VOID_TYPE) (void) (*fv)(p1, p2, p3, p4, p5);
#ifdef FLOAT_TYPE
	  else if (type == FLOAT_TYPE) xf = (*ff)(p1, p2, p3, p4, p5);
#endif
	  else ret = (*f)(p1, p2, p3, p4, p5);
	  break;

      default: 
	lang_doerror("Function requires too many parameters");
	Lang_Error = UNKNOWN_ERROR;
	break;
     }

   if (!Lang_Error) switch(type)
     {
      case STRING_TYPE: lang_push_string((char *) ret); break;
      case INT_TYPE:
	/* For msdos, longs are 4 bytes and ints are two.  Take this
	   approach: */
	lang_push_integer(*(int*) &ret); break;
      case VOID_TYPE: break;
#ifdef FLOAT_TYPE
      case FLOAT_TYPE: lang_push_float(* (float *) &xf); break;
#endif
      default: Lang_Error = TYPE_MISMATCH;
     }
   /* I free afterword because functions that return char * may point to this
      space. */
   if (((p5type & 0xFF) == LANG_DATA) && ((p5type >> 8) == STRING_TYPE)) FREE(p5);
   if (((p4type & 0xFF) == LANG_DATA) && ((p4type >> 8) == STRING_TYPE)) FREE(p4);
   if (((p1type & 0xFF) == LANG_DATA) && ((p1type >> 8) == STRING_TYPE)) FREE(p1);
   if (((p2type & 0xFF) == LANG_DATA) && ((p2type >> 8) == STRING_TYPE)) FREE(p2);
   if (((p3type & 0xFF) == LANG_DATA) && ((p3type >> 8) == STRING_TYPE)) FREE(p3);
}

void lang_do_loops(unsigned char type, long addr)
{
   register int i, ctrl;
   int ctrl1;
   int first, last, one = 0;
   register Lang_Object_Type *obj1, *obj2, *block;

   block = (Lang_Object_Type *) addr;
   obj1 = (Lang_Object_Type *) block->value;

   switch (type)
     {
      case LANG_WHILE:

	/* we need 2 blocks: first is the control, the second is code */
	block++;
	if ((block->type) != LANG_BLOCK)
	  {
	     lang_doerror("Block needed for while.");
	     return;
	  }
	obj2 = (Lang_Object_Type *) block->value;

	while(!Lang_Error)
	  {
	     inner_interp(obj1);
	     if (Lang_Break) break;
	     if (Lang_Error || !lang_pop_integer(&ctrl1)) return;
	     if (!ctrl1) break;
	     inner_interp(obj2);
	     if (Lang_Break) break;
	     Lang_Break_Condition = Lang_Continue = 0;
	  }
	break;

      case LANG_FOR:  /* 3 elements: first, last, step */
	if (!lang_pop_integer(&ctrl1)) return;
	if (!lang_pop_integer(&last)) return;
	if (!lang_pop_integer(&first)) return;
	ctrl = ctrl1;
	if (ctrl >= 0)
	  {	     
	     for (i = first; i <= last; i += ctrl)
	       {
		  if (Lang_Error) return;
		  lang_push_integer(i);
		  inner_interp(obj1);
		  if (Lang_Break) break;
		  Lang_Break_Condition = Lang_Continue = 0;
	       }
	  }
	else
	  {
	     for (i = first; i >= last; i += ctrl)
	       {
		  if (Lang_Error) return;
		  lang_push_integer(i);
		  inner_interp(obj1);
		  if (Lang_Break) break;
		  Lang_Break_Condition = Lang_Continue = 0;
	       }
	  }
	
	break;

      case LANG_LOOP:
	if (!lang_pop_integer(&ctrl1)) return;
	ctrl = ctrl1;
      case LANG_FOREVER:
	if (type == LANG_FOREVER) one = 1;
	while (one || (ctrl-- > 0))
	  {
	     if (Lang_Error) break;
	     inner_interp(obj1);
	     if (Lang_Break) break;
	     Lang_Break_Condition = Lang_Continue = 0;
	  }
	break;

      default:  lang_doerror("Unknown loop type.");
     }
   Lang_Break = Lang_Continue = 0;
   Lang_Break_Condition = Lang_Return;
}

void lang_do_ifs(Lang_Object_Type *addr)
{
   unsigned char type;
   long val;
   int test, doit;

   type = addr->type >> 8;
   if (!lang_pop_integer(&test)) return;

   addr--;

   doit = 1;         /* optimistic choice */
   switch (type)
     {
	case LANG_IF:  if (!test) doit = 0; break;
	case LANG_IFNOT: if (test) doit = 0; break;
	case LANG_ELSE: if (test) addr--; break;
	default:
	  lang_doerror("Unknown IF type!");
     }

   if (addr->type != LANG_BLOCK)  /* was & 0xFF as well */
     {
	lang_doerror("Block needed.");
	return;
     }

   if (doit)
     {
	val = addr->value;
	inner_interp((Lang_Object_Type *) val);
     }
}

void lang_do_else(unsigned char type, long addr)
{
   long val;
   int test, status;
   char *str = NULL;
   Lang_Object_Type *obj, cobj;

   obj = (Lang_Object_Type *) addr;

   if (type == LANG_SWITCH)
     {
	if (!lang_pop(&cobj)) return;
	if (IS_DATA_STRING(cobj)) str = (char *) cobj.value;
     }

   while((obj->type == LANG_BLOCK) != 0)
     {
	val = obj->value;
	if (type == LANG_SWITCH)
	  {
	     if (str == NULL) lang_push(&cobj); else lang_push_string(str);
	  }

	status = inner_interp((Lang_Object_Type *) val);
	if (Lang_Error || Lang_Break_Condition) return;
	if (type == LANG_SWITCH)
	  {
	     if (status) break;
	  }

	else if (!lang_pop_integer(&test)) return;
	if (((type == LANG_ANDELSE) && !test)
	    || ((type == LANG_ORELSE) && test)) break;
	obj++;
     }
   if (type != LANG_SWITCH) lang_push_integer(test);
   else if (str != NULL) FREE(str);
   return;
}

void lang_dump(char *s)
{
   fputs(s, stderr);
}

void (*Lang_Dump_Routine)(char *) = lang_dump;

extern void do_traceback(Lang_Name_Type *nt, int locals);

/* A comment */
static void execute_function(Lang_Name_Type *entry)
{
   int n_locals, i;
   long val;
   Lang_Object_Type *frame;
   
   n_locals = (entry->obj.type) >> 8;

   /* need loaded?  */
   if (n_locals == 255)
     {
	if (!lang_load_file((char *) entry->obj.value)) return;
	n_locals = (entry->obj.type) >> 8;
	if (n_locals == 255)
	  {
	     lang_doerror("Function did not autoload!");
	  }
     }
   
   /* let the lang error propagate through since it will do no harm
      and allow us to restore stack. */
   val = entry->obj.value;
   /* set new stack frame */
   frame = Local_Variable_Frame;
   i = n_locals;
   while(i--)
     {
	Local_Variable_Frame++;
	if (Local_Variable_Frame - Local_Variable_Stack > MAX_LOCAL_STACK)
	  {
	     lang_doerror("Local Variable Stack Overflow!");
	  }
	Local_Variable_Frame->type = 0;
     }

   if (Lang_Enter_Function != NULL) (*Lang_Enter_Function)(entry->name + 1);
   if (!Lang_Error) inner_interp((Lang_Object_Type *) val);
   if (Lang_Exit_Function != NULL) (*Lang_Exit_Function)(entry->name + 1);
   
   if (Lang_Error && Lang_Traceback)
     {
	do_traceback(entry, n_locals);
     }
   /* free local variables.... */
   while(Local_Variable_Frame > frame)
     {
	if (IS_DATA_STRING(*Local_Variable_Frame))
	FREE (Local_Variable_Frame->value);
	Local_Variable_Frame--;
     }
   Lang_Break_Condition = Lang_Return = Lang_Break = 0;
}
void do_traceback(Lang_Name_Type *nt, int locals)
{
   char buf[80];
   int i;
   Lang_Object_Type *objp;
   long val;
   unsigned short stype;
   
   sprintf(buf, "S-Lang Traceback: %s\n",nt->name + 1);
   (*Lang_Dump_Routine)(buf);
   if (!locals) return;
   (*Lang_Dump_Routine)("  Local Variables:\n");
   
   for (i = 0; i < locals; i++)
     {
	objp = Local_Variable_Frame - i;
	stype = objp->type >> 8;
	val = objp->value;
	
	if (STRING_TYPE == stype)
	  {
	     sprintf(buf, "\t$%d: \"", i);
	     (*Lang_Dump_Routine)(buf);
	     (*Lang_Dump_Routine)((char *) val);
	     (*Lang_Dump_Routine)("\"\n");
	     continue;
	  }
	else if (INT_TYPE == stype)
	  {
	     sprintf(buf, "\t$%d: %d\n", i, (int) val);
	  }
	else sprintf(buf, "\t$%d: ??\n", i);
	(*Lang_Dump_Routine)(buf);
     }
}

static void call_funptr(Lang_Name_Type *optr)
{
   Lang_Object_Type objs[2];
   
   if (optr == NULL)
     {
	lang_doerror("Object Ptr is Nil!");
	return;
     }
   
   objs[0].value = (long) optr;
   objs[0].type = optr->obj.type;
   objs[1].type = 0;
   objs[1].value = 0;
   inner_interp(objs);
}

   



/* inner interpreter */
int inner_interp(Lang_Object_Type *addr)
{
   Lang_Object_Type obj1, obj2;
   register unsigned short type;
   register unsigned char stype;
   int x, y, z;
   register long val;
   long block = 0;
   Lang_Name_Type *name_entry;
   Lang_Object_Type *err_block = NULL, *objp;
   /* if non-null, execute this on error */
   int save_err;

#ifdef FLOAT_TYPE
   float xf, yf, zf;
#endif

   if (addr == NULL)
     {
	Lang_Error = UNKNOWN_ERROR;
	return(0);
     }
   
   while((type = addr->type) != 0)
     {
	if (Lang_Break_Condition)
	  {
	     if (Lang_Return || Lang_Break)
	       {
		  Lang_Break = 1;
		  return(1);
	       }
	     if (Lang_Continue) return(1);
	  }
	
	val = addr->value;
	stype = type >> 8;

	switch (type & 0xFF)
	  {
	   case LANG_INTRINSIC:
	     name_entry = (Lang_Name_Type *) val;
	     lang_do_intrinsic((Lang_Object_Type *) &name_entry->obj);
	     if (Lang_Error && Lang_Traceback)
	       {
		  do_traceback(name_entry, 0);
	       }
	     break;
	   case LANG_BLOCK:
	     if (stype == ERROR_BLOCK) err_block = addr;
	     else if (!block) block = (long) addr;
	     break;  /* don not know what to do yet */
	     
	   case LANG_FUNCTION:
	     /* val is a reference to the name table, so get its obj: */
	     name_entry = (Lang_Name_Type *) val;
	     execute_function(name_entry);
	     break;

           case LANG_LVARIABLE:      /* make val point to local stack */
	       val = (long) (Local_Variable_Frame - (int) val);
	       lang_push_variable((Lang_Object_Type *) val);
	       break;
	   case LANG_GVARIABLE: 
	       val = ((Lang_Name_Type *) val)->obj.value;
	       lang_push_variable((Lang_Object_Type *) val);
	       break;
	     
             case LANG_LITERAL:        /* a constant */
	       lang_push(addr);
	       break;

	   case LANG_DIRECTIVE: 
	     if (stype & LANG_EQS_MASK)
	       {
		  lang_do_eqs(addr);
		  break;
	       }

	     if (!block) lang_doerror("No Blocks!");
	     else if (stype & LANG_LOOP_MASK) lang_do_loops(stype, block);
	     else if (stype & LANG_IF_MASK) lang_do_ifs(addr);
	     else if (stype & LANG_ELSE_MASK) lang_do_else(stype, block);
	     /* else lang_doerror("Unknown directive!"); */
	     block = 0;
	     break;
	     
	     case LANG_IVARIABLE:
	     case LANG_RVARIABLE:
	       val = ((Lang_Name_Type *) val)->obj.value;
	       switch(stype)
	       {
		  case STRING_TYPE: lang_push_string((char *) val); break;
		  case INT_TYPE: lang_push_integer(*(int *) val); break;
		  default: lang_doerror("Unsupported Type!");
	       }

	       break;

	   case LANG_CMP: 
	   case LANG_BINARY: 
	       z = 0;
#ifndef FLOAT_TYPE
	       if (!lang_pop_integer(&y) || !lang_pop_integer(&x)) return(0);
	       switch (stype)
		 {
		    case LANG_EQ: if (x == y) z = 1; break;
		    case LANG_NE: if (x != y) z = 1; break;
		    case LANG_GT: if (x > y) z = 1; break;
		    case LANG_GE: if (x >= y) z = 1; break;
		    case LANG_LT: if (x < y) z = 1; break;
		    case LANG_LE: if (x <= y) z = 1; break;
		    case LANG_OR: if (x || y) z = 1; break;
		    case LANG_AND: if (x && y) z = 1; break;
		    case LANG_BAND: z = x & y; break;
		    case LANG_BXOR: z = x ^ y; break;
		    case LANG_BOR: z = x | y; break;
		    case LANG_PLUS: z = x + y; break;
		    case LANG_MINUS: z = x - y; break;
		    case LANG_TIMES: z = x * y; break;
		    case LANG_DIVIDE: z = x / y; break;   /* y == 0? */
		    case LANG_SHL: z = x << y; break;
		    case LANG_SHR: z = x >> y; break;
		    default:
		      Lang_Error = INTERNAL_ERROR;
		      return(0);
		   }
		 lang_push_integer(z);
		 break;			/* binary */
#else /* FLOAT_TYPE */
	       if (!lang_pop_float(&yf) || !lang_pop_float(&xf)) return(0);
	       switch (stype)
		 {
		    /* case LANG_SHR, LANG_SHL needs done */
		    case LANG_EQ: if (xf == yf) z = 1; break;
		    case LANG_NE: if (xf != yf) z = 1; break;
		    case LANG_GT: if (xf > yf) z = 1; break;
		    case LANG_GE: if (xf >= yf) z = 1; break;
		    case LANG_LT: if (xf < yf) z = 1; break;
		    case LANG_LE: if (xf <= yf) z = 1; break;
		    case LANG_OR: if (xf || yf) z = 1; break;
		    case LANG_AND: if (xf && yf) z = 1; break;
		    case LANG_PLUS: zf = xf + yf; break;
		    case LANG_MINUS: zf = xf - yf; break;
		    case LANG_TIMES: zf = xf * yf; break;
		    case LANG_DIVIDE: zf = xf / yf; break;   /* y == 0? */
		    default:
		      Lang_Error = INTERNAL_ERROR;
		      return(0);
		   }
		 if (type == LANG_CMP) lang_push_integer(z);
		 else lang_push_float(zf);
		 break;			/* binary */
#endif /* float */
	   case LANG_RETURN: 
	     Lang_Break_Condition = Lang_Return = Lang_Break = 1; return(1);
	   case LANG_BREAK: 
	     Lang_Break_Condition = Lang_Break = 1; return(1);
	   case LANG_CONTINUE: 
	     Lang_Break_Condition = Lang_Continue = 1; return(1);
	     
	   case LANG_EXCH: if (!lang_pop(&obj1) || !lang_pop(&obj2)) return(1);
	     lang_push(&obj1); lang_push(&obj2);
	     break;

	   case LANG_LABEL:
	     if (!lang_pop_integer(&z) || !z) return(0);
	     break;

	   case LANG_UNARY:
#ifndef FLOAT_TYPE
	     if (!lang_pop_integer(&z)) return(0);
	     switch (stype)
	       {
		  case LANG_NOT:  z = !z; break;
		  case LANG_BNOT:  z = ~z; break;
		  case LANG_CHS:  z = -z; break;
		case LANG_ABS: z = abs(z); break;
		case LANG_SIGN: z = (z >= 0) ? 1 :  -1; break;
		  default: Lang_Error = INTERNAL_ERROR; return(0);
	       }
	     lang_push_integer(z);
	     break;			/* unary */
#else
	     if (stype == LANG_CHS)
	       {
		  if (!lang_pop_float(&zf)) return(0);
		  zf = -zf;
		  lang_push_float(zf);
		  break;
	       }
	     else
	       {
		  if (! lang_pop_integer(&z)) return(0);
		  if (stype == LANG_NOT) z = !z;
		  else if (stype == LANG_BNOT) z = ~z;
		  else
		    {
		       Lang_Error = INTERNAL_ERROR;
		       return(0);
		    }
		  lang_push_integer(z);
	       }

	     break;
#endif
	   case LANG_LOBJPTR:
	     val = (long) (Local_Variable_Frame - (int) val);
	     objp = (Lang_Object_Type *) val;
	     if (objp->type == 0)
	       {
		  lang_doerror("Local variable pointer not initialized.");
		  break;
	       }
	     
	     obj1.value = objp->value;
	     obj1.type = LANG_DATA | (LANG_OBJ_TYPE << 8);
	     lang_push(&obj1);
	     break;
	     
	   case LANG_GOBJPTR:
	     obj1.value = val;
	     obj1.type = LANG_DATA | (LANG_OBJ_TYPE << 8);
	     lang_push(&obj1);
	     break;
	     	     
	   case LANG_X_ERROR: 
	     if (err_block != NULL) 
	       {
		  inner_interp((Lang_Object_Type *)(err_block->value));
		  if (Lang_Error) err_block = NULL;
	       }
	     else lang_doerror("No Error Block");
	     break;
	     default: lang_doerror("Run time error.");
	  }
	if (Lang_Error) break;
	addr++;
     }
   if ((Lang_Error) && (err_block != NULL) && 
       ((Lang_Error == USER_BREAK) || (Lang_Error == INTRINSIC_ERROR)))
     {
	save_err = Lang_Error;
	Lang_Error = 0;
	inner_interp((Lang_Object_Type *)(err_block->value));
	Lang_Error = save_err;
     }
   
   return(1);
}

/* Hash value of current item to search in table */
static unsigned char Hash;

/* offsets into intrinsic table */
static int Name_Index[256] =
{
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
   -1,-1,-1,-1,-1,-1
};


static void compute_hash(unsigned char *s)
{
   register unsigned char *ss = s;
   Hash = 0;
   while (*ss) Hash += *ss++ + (Hash << 5);
   if (!Hash) Hash = *s;
}


Lang_Name_Type *lang_locate_name_in_table(char *name, Lang_Name_Type *table, Lang_Name_Type *t0, int max)
{
   register char *nm;
   register Lang_Name_Type *t = t0, *tmax = table + max;
   char h = Hash;
   
   while((t != tmax) && (nm = t->name, *nm != 0))
     {
	if ((*nm == h) && !strcmp(nm + 1,name))
	  {
#ifdef SLANG_STATS
	     t->n++;
#endif
	     return(t);
	  }
	t++;
     }
   if ((table == Lang_Intrinsic_Name_Table) || (t == tmax)) return(NULL);   
   return(t);
}

#ifdef SLANG_STATS
int lang_dump_stats(char *file)
{
   Lang_Name_Type *t = Lang_Intrinsic_Name_Table;
   FILE *fp;
   if ((fp = fopen(file, "w")) == NULL) return(0);
   while (*t->name != 0)
     {
	fprintf(fp, "%3d\t%3d\t%s\n", t->n, (int) (unsigned char) *t->name, t->name + 1);
	t++;
     }
   fclose(fp);
   return(1);
}
#endif

Lang_Name_Type *lang_locate_name(char *name)
{
   unsigned char *nm;
   Lang_Name_Type *table;
   
   Hash = 0; 
   nm = (unsigned char *) name;
   while (*nm) Hash += (Hash << 5) + *nm++;
   if (!Hash) Hash = (unsigned char) *name;
   
   table = Lang_Local_Variable_Table;

   if (table != NULL) 
     table = lang_locate_name_in_table(name, table, table, MAX_LOCAL_VARIABLES);
   if ((table == NULL) || (*table->name == 0))
     {
	table = Lang_Intrinsic_Name_Table;
	table = lang_locate_name_in_table(name, table, table + Name_Index[Hash], -1);
	if ((table == NULL) || (*table->name == 0))
	  {
	     table = lang_locate_name_in_table(name, Lang_Name_Table, Lang_Name_Table, LANG_MAX_SYMBOLS);
	  }
     }
   return(table);
}


void lang_array_sort(char *f)
{
   Array_Type *at_str, *at_int;
   unsigned char type;
   Lang_Name_Type *entry;
   Lang_Object_Type obj;
   int l, j, ir, i, rra, n, cmp;
   int *ra;
   long ptr;
   
   
   if ((NULL == (entry = lang_locate_name(f))) || (*entry->name == 0))
     {
	lang_doerror("Sort function undefined.");
	return;
     }
   
   type = entry->obj.type & 0xFF;
   if (type != LANG_FUNCTION)
     {
	lang_doerror("Invalid sort function.");
	return;
     }
   
   if (NULL == (at_str = pop_array())) return;
   
   n = at_str->x;
   
   if (at_str->dim != 1)
     {
	lang_doerror("Sort requires 1 dim arrays.");
	return;
     }
   
   ptr = (long) NULL;
   if (NULL == (at_int = lang_create_array(&ptr, 1, n, 1, 1, "i")))
     {
	lang_doerror("Error Creating index array.");
	return;
     }
   
   ra = (int *) at_int->ptr;
   ra--;
   for (i = 1; i <= n; i++) ra[i] = i;
   
   /* heap sort from adapted from numerical recipes */
   
   l = 1 + n / 2;
   ir = n;
   
   while(1)
     {
	if (l > 1) rra = ra[--l];
	else
	  {
	     rra = ra[ir];
	     ra[ir] = ra[1];
	     if (--ir == 1)
	       {
		  ra[1] = rra;
		  obj.type = LANG_DATA | (ARRAY_TYPE << 8);
		  obj.value = (long) at_int;
		  lang_push(&obj);
		  return;
	       }
	  }
	i = l;
	j = 2 * l;
	while(j <= ir)
	  {
	     if (j < ir)
	       {
		  array_push_element(at_str, ra[j] - 1);
		  array_push_element(at_str, ra[j + 1] - 1);
		  execute_function(entry);
		  if (!lang_pop_integer(&cmp)) return;
		  if (cmp) j++;
	       }
	     array_push_element(at_str, rra - 1);
	     array_push_element(at_str, ra[j] - 1);
	     execute_function(entry);
	     if (!lang_pop_integer(&cmp)) return;
	     
	     if (cmp) 
	       {
		  ra[i] = ra[j];
		  i = j;
		  j += j;
	       }
	     else j = ir + 1;
	  }
	ra[i] = rra;
     }
}


void lang_add_name(char *name, long addr, unsigned short type)
{
   Lang_Name_Type *entry;
   unsigned char stype;
   char *p, ch, *err = "Name Syntax";
   
   /* check syntax.  Allowed chars are: $!_?AB..Zab..z0-9 */
   p = name;
   while ((ch = *p++) != 0)
     {
	if ((ch >= 'a') && (ch <= 'z')) continue;
	if ((ch >= 'A') && (ch <= 'Z')) continue;
	if ((ch >= '0') && (ch <= '9')) continue;
	if ((ch == '_') || (ch == '$') || (ch == '!') || (ch == '?')) continue;
	lang_doerror(err);
	return;
     }

   if (strlen(name) > LANG_MAX_NAME_LEN)
     {
	lang_doerror("Name too long.");
	return;
     }

   
   if (NULL == (entry = lang_locate_name(name)))
     {
	lang_doerror("Table size exceeded!");
	return;  /* table full */
     }
   
   if (*entry->name != 0)
     {
	stype = entry->obj.type & 0xFF;
#if 0
	if ((stype != LANG_INTRINSIC) && (stype != LANG_GVARIABLE)
	    && (stype != LANG_IVARIABLE) && (stype != LANG_RVARIABLE))
	  {
	     /* free its function or block definition only if not autoloaded */
	     if (! ((stype == LANG_FUNCTION) && ((entry->obj.type >> 8) == 255)))
	       lang_free_branch((Lang_Object_Type *) entry->obj.value);

	     FREE(entry->obj.value);
	  }
#endif
	/* 255 denotes that the function needs autoloaded. */
	if (stype == LANG_FUNCTION)
	  {
	     if ((entry->obj.type >> 8) != 255)
	       lang_free_branch((Lang_Object_Type *) entry->obj.value);
	     FREE(entry->obj.value);
	  }
     }
   else 
     {
	strcpy(entry->name + 1, name);
	*entry->name = (char) Hash;
     }
   

   entry->obj.value = (long) addr;
   entry->obj.type = type;
}

void lang_autoload(char *name, char *file)
{
   unsigned short type;
   long f;

   type = LANG_FUNCTION | (255 << 8);
   f = lang_make_string(file);

   lang_add_name(name, f, type);
}

void lang_define_function(char *name)
{
   long addr;
   unsigned short type;

   Lang_Defining_Function = 0;
   addr = (long) Lang_Function_Body;
   type = LANG_FUNCTION | (Local_Variable_Number << 8);
   if (name != NULL) lang_add_name(name, addr, type);
   if (Lang_Error) return;
   if (Lang_Local_Variable_Table != NULL) FREE(Lang_Local_Variable_Table);
   Lang_Local_Variable_Table = NULL;
   Local_Variable_Number = 0;

   /* terminate function */
   Lang_Object_Ptr->type = 0;
   Lang_Object_Ptr = Lang_Interp_Stack_Ptr;   /* restore pointer */
}

/* call inner interpreter or return for more */
void lang_try_now(void)
{
   Lang_Object_Type *old_stack, *old_stack_ptr, *old_int_stack_ptr;
   Lang_Object_Type new_stack[4];
   int i;

   if (Lang_Defining_Function || Lang_Defining_Block)
     {
	Lang_Object_Ptr++;
	return;
     }

   /* This is the entry point into the inner interpreter.  As a result, it
      is also the exit point of the inner interpreter.  So it is necessary to
      clean up if there was an error.
    */

   (Lang_Object_Ptr + 1)->type = 0;  /* so next command stops after this */

   /* now before entering the inner interpreter, we make a new stack so that
      we are able to be reentrant */
   
   for (i = 1; i < 4; i++)
     {
	new_stack[i].type = 0;
	new_stack[i].value = 0;
     }
   
   /* remember these values */
   old_int_stack_ptr = Lang_Interp_Stack_Ptr;
   old_stack_ptr = Lang_Object_Ptr;
   old_stack = Lang_Interp_Stack;

   /* new values for reentrancy */
   Lang_Interp_Stack_Ptr = Lang_Object_Ptr = Lang_Interp_Stack = new_stack;

   /* now do it */
   inner_interp(old_stack);

   /* we are back so restore old pointers */
   Lang_Interp_Stack_Ptr = old_int_stack_ptr;
   Lang_Object_Ptr = old_stack_ptr;
   Lang_Interp_Stack = old_stack;

   /* now free blocks from the current interp_stack.  There can only 
      be blocks since they are only objects not evaluated immediately */

   while(Lang_Object_Ptr != Lang_Interp_Stack)
     {
	/* note that top object is not freed since it was not malloced */
	Lang_Object_Ptr--;
	/* FREE(Lang_Object_Ptr->value); */
	lang_free_branch((Lang_Object_Type *) Lang_Object_Ptr->value);
     }

   /* now free up the callocd stack. 
   FREE(new_stack); */
}


#define eqs(a,b) ((*(a) == *(b)) && !strcmp(a,b))
int lang_execute_function(char *name)
{
   unsigned char type;
   Lang_Name_Type *entry;
   if ((NULL == (entry = lang_locate_name(name))) || (*entry->name == 0)) return(0);
   type = entry->obj.type & 0xFF;
   if (type == LANG_FUNCTION) execute_function(entry);
   else if (type == LANG_INTRINSIC) 
     lang_do_intrinsic((Lang_Object_Type *) &(entry->obj));
   else return(0);
   if (Lang_Error) lang_doerror(NULL);
   return(1);
}

#if 0
static void make_name_ptr(char *name)
{
   Lang_Name_Type *n;
   Lang_Object_Type obj;
   
   n = lang_locate_name(name);
   
   if ((n == NULL) || (*n->name == 0))
     {
	lang_doerror("Object is undefined.");
	return;
     }
   
   obj.type = LANG_DATA | (LANG_OBJ_TYPE << 8);
   
   if ((n->obj.type >> 8) == LANG_OBJ_TYPE) obj.value = n->obj.value;
   else obj.value = (long) n;
   lang_push (&obj);
}
#endif

static int lang_exec(char *name)
{
   Lang_Name_Type *entry;
   short type;
   int ptr_type = 0;
   
   
   if (eqs(name, "EXECUTE_ERROR_BLOCK"))
     {
	Lang_Object_Ptr->type = LANG_X_ERROR;
	Lang_Object_Ptr->value = 0;
     }
   else
     {
	if (*name == '&')
	  {
	     name++;
	     ptr_type = 1;
	  }
	
	if (NULL == (entry = lang_locate_name(name))) return(0);
	if (*entry->name == 0) return(0); 
		     
	type = entry->obj.type;
	if (ptr_type)
	  {
	     Lang_Object_Ptr->type = type == LANG_LVARIABLE ?
	        LANG_LOBJPTR : LANG_GOBJPTR;
	  }
	else
	  {
	     Lang_Object_Ptr->type = type;
	  }
	
	if (type == LANG_LVARIABLE)
	  {
	     Lang_Object_Ptr->value = entry->obj.value;
	  }
	else
	  {
	     Lang_Object_Ptr->value = (long) entry;
	  }
     }
   
   lang_try_now();
   return(1);
}

/* arrays from these must be in alphabetic order!! */
typedef struct Lang_Name2_Type
{
   char *name;  int type;
} Lang_Name2_Type;


static int eqs_name(char *t, Lang_Name2_Type *d)
{
   register char *p;
   while ((p = d->name) != NULL)
     {
	if ((*t == *p) && !strcmp(t, p)) return(d->type);
	d++;
     }
   return(0);
}

static Lang_Name2_Type Lang_Binaries[] = 
{
   {"!=", LANG_NE},
   {"&", LANG_BAND},
   {"*", -LANG_TIMES},
   {"+", -LANG_PLUS},
   {"-", -LANG_MINUS},
   {"/", -LANG_DIVIDE},
   {"<", LANG_LT},
   {"<=", LANG_LE},
   {"==", LANG_EQ},
   {">", LANG_GT},
   {">=", LANG_GE},
   {"and", LANG_AND},
   {"or", LANG_OR},
   {"shl", LANG_SHL},
   {"shr", LANG_SHR},
   {"xor", LANG_BXOR},
   {"|", LANG_BOR},
   {NULL, NULL}
};


int lang_try_binary(char *t)
{
   int ssub;
   unsigned char sub, type;
   ssub = 0;

   if (0 == (ssub = eqs_name(t, Lang_Binaries))) return(0);

   if (ssub < 0)
     {
	ssub = -ssub;
	type = LANG_BINARY;
     }
   else type = LANG_CMP;
   sub = (unsigned char) ssub;

   Lang_Object_Ptr->type = type | (sub << 8);
   Lang_Object_Ptr->value = 0;         /* not used */

   lang_try_now();
   return(1);
}

int lang_try_unary(char *t)
{
   unsigned char ssub, type;

   if (eqs(t, "~")) ssub = LANG_BNOT;
   else if (eqs(t, "not")) ssub = LANG_NOT;
   else if (eqs(t, "chs")) ssub = LANG_CHS;
   else if (eqs(t, "sign")) ssub = LANG_SIGN;
   else if (eqs(t, "abs")) ssub = LANG_ABS;
   else return(0);

   type = LANG_UNARY;

   Lang_Object_Ptr->type = type | (ssub << 8);
   Lang_Object_Ptr->value = 0;         /* not used */

   lang_try_now();
   return(1);
}

void lang_begin_function(void)
{
   if (Lang_Defining_Function || Lang_Defining_Block)
     {
	lang_doerror("Function nesting illegal.");
	return;
     }

   Lang_Defining_Function = 1;

   /* make initial size for 3 things */
   Lang_FBody_Size = 3; 
   if (NULL == (Lang_Function_Body = (Lang_Object_Type *)
          CALLOC(Lang_FBody_Size, sizeof(Lang_Object_Type))))
     {
	lang_doerror("Calloc error defining function.");
	return;
     }
   /* function definitions should be done only at top level so it should be
      safe to do this: */
   Lang_Interp_Stack_Ptr = Lang_Object_Ptr;
   Lang_Object_Ptr = Lang_FBody_Ptr = Lang_Function_Body;
   return;
}

void lang_end_block(void)
{
   Lang_Object_Type *node, *branch;
   Lang_Block_Depth--;

   /* terminate the block */
   Lang_Object_Ptr->type = 0;

   branch = Lang_Block_Body;

   if (Lang_Block_Depth == -1)         /* done */
     {
	if (Lang_Defining_Function)
	  {
	     node = Lang_FBody_Ptr++;
	  }
	else node = Lang_Interp_Stack_Ptr;   /* on small stack */
     }
   else                                /* pop previous block */
     {
	Lang_BBody_Size = Lang_Block_Stack[Lang_Block_Depth].size;
	Lang_Block_Body = Lang_Block_Stack[Lang_Block_Depth].body;
	node = Lang_Block_Stack[Lang_Block_Depth].ptr;
     }

   node->type = LANG_BLOCK;
   node->value = (long) branch;
   Lang_Object_Ptr = node + 1;
   Lang_Defining_Block--;
}

void lang_begin_block(void)
{
   if (Lang_Block_Depth == LANG_MAX_BLOCKS - 1)
     {
	lang_doerror("Block Nesting too deep.");
	Lang_Error = UNKNOWN_ERROR;
	return;
     }
   /* push the current block onto the stack */
   if (Lang_Block_Depth > -1)
     {
	Lang_Block_Stack[Lang_Block_Depth].size = Lang_BBody_Size;
	Lang_Block_Stack[Lang_Block_Depth].body = Lang_Block_Body;
	Lang_Block_Stack[Lang_Block_Depth].ptr = Lang_Object_Ptr;
     }

   /* otherwise this is first block so save function pointer */
   else if (Lang_Defining_Function) Lang_FBody_Ptr = Lang_Object_Ptr;
   else Lang_Interp_Stack_Ptr = Lang_Object_Ptr;

   Lang_BBody_Size = 5;    /* 40 bytes */
   if (NULL == (Lang_Block_Body = (Lang_Object_Type *)
                   CALLOC(Lang_BBody_Size, sizeof(Lang_Object_Type))))
      {
	 Lang_Error = MALLOC_ERROR;
	 /* lang_doerror("Malloc error defining block."); */
	 return;
      }
   Lang_Block_Depth++;
   Lang_Defining_Block++;
   Lang_Object_Ptr = Lang_Block_Body;
   return;
}


/* see if token is a directive, and add it to current block/function */
static Lang_Name2_Type Lang_Directives[] =
{
   {"!if", LANG_IFNOT},
   {"andelse", LANG_ANDELSE},
   {"else", LANG_ELSE},
   {"for", LANG_FOR},
   {"forever", LANG_FOREVER},
   {"if", LANG_IF},
   {"loop", LANG_LOOP},
   {"orelse", LANG_ORELSE},
   {"switch", LANG_SWITCH},
   {"while", LANG_WHILE},
   {NULL, NULL}   
};

static int try_directive(char *t, int *flag)
{  
   unsigned char sub = 0;
   unsigned short type = LANG_DIRECTIVE;
   Lang_Object_Type *lop;
   int flag_save;
   
   if ((sub = (unsigned char) eqs_name(t, Lang_Directives)) != 0); /* null */
   else if (*flag && eqs(t, "ERROR_BLOCK"))
     {
	lop = Lang_Object_Ptr - 1;
	if (lop->type != LANG_BLOCK) lang_doerror("Internal Error with error_block!");
	else lop->type = LANG_BLOCK | (ERROR_BLOCK << 8);
	return(1);
     }
   /* rest valid only if flag is zero */
   else if (*flag) return(0);
   else
     {
	if (Lang_Defining_Block && eqs(t, "continue")) type = LANG_CONTINUE;
	else if (Lang_Defining_Block && eqs(t, "break")) type = LANG_BREAK;
	else if (Lang_Defining_Function && eqs(t, "return")) type = LANG_RETURN;
	/* why is exch here? */
	else if (eqs(t, "exch")) type = LANG_EXCH;
	else return(0);
	*flag = 1;
     }

   Lang_Object_Ptr->type = type | (sub << 8);
   Lang_Object_Ptr->value = 0;         /* not used */

   flag_save = *flag; *flag = 0;
   lang_try_now();
   *flag = flag_save;

   return(1);
}

long lang_make_object(void)
{
   Lang_Object_Type *obj;

   obj = (Lang_Object_Type *) MALLOC(sizeof(Lang_Object_Type));
   if (NULL == obj)
     {
	Lang_Error = MALLOC_ERROR; /* lang_doerror("Lang: malloc error."); */
	return(0);
     }
   obj->type = 0;
   obj->value = 0;
   return (long) obj;
}

int interp_variable_eqs(char *name)
{
   Lang_Name_Type *v;
   Lang_Object_Type obj;
   unsigned short type;
   unsigned char stype;
   char ch;
   long value;
   int offset;
   int eq, pe, me, pp, mm;
   
   eq = LANG_GEQS - LANG_GEQS;
   pe = LANG_GPEQS - LANG_GEQS;
   me = LANG_GMEQS - LANG_GEQS;
   pp = LANG_GPP - LANG_GEQS;
   mm = LANG_GMM - LANG_GEQS;

   /* Name must be prefixed by one of:  =, ++, --, +=, -= 
      all of which have ascii codes less than or equal to 61 ('=') */
   
   offset = -1;
   ch = *name++;
   switch (ch)
     {
      case '=': offset = eq; break;
      case '+': 
	ch = *name++;
	if (ch == '+') offset = pp; else if (ch == '=') offset = pe;
	break;
      case '-':
	ch = *name++;
	if (ch == '-') offset = mm; else if (ch == '=') offset = me;
	break;
     }
   
   if (offset == -1) return 0;
   
   v = lang_locate_name(name);
   if ((v == NULL) || *(v->name) == 0)
     {
	Lang_Error = UNDEFINED_NAME;
	return(1);
     }

   type = (v->obj.type) & 0xFF;
   if (type == LANG_RVARIABLE)
     {
	Lang_Error = READONLY_ERROR;
	return(1);
     }

   if ((type != LANG_GVARIABLE) && (type != LANG_LVARIABLE)
       && (type != LANG_IVARIABLE))
     {
	Lang_Error = DUPLICATE_DEFINITION;
	return(1);
     }

   /* its value is location of object in name table */
   value = (long) v;

   if (type == LANG_IVARIABLE)
     {
	if ((v->obj.type >> 8) == STRING_TYPE)
	  {
	     Lang_Error = READONLY_ERROR;
	     return(1);
	  }

	stype = LANG_IEQS;
     }

   else if (type == LANG_GVARIABLE) stype = LANG_GEQS;
   else
     {
	stype = LANG_LEQS;
	value = v->obj.value;
     }

   stype += offset;
   
   if (Lang_Defining_Function || Lang_Defining_Block)
     {
	Lang_Object_Ptr->type = LANG_DIRECTIVE | (stype << 8);
	Lang_Object_Ptr->value = value;
	Lang_Object_Ptr++;
	return (1);
     }

   /* create an object with the required properties for next call */
   obj.type = LANG_DIRECTIVE | (stype << 8);
   obj.value = value;
   lang_do_eqs(&obj);
   return(1);
}

unsigned char is_number(char *t)
{
   char *p;

   p = t;
   if (*p == '-') p++;
   while ((*p >= '0') && (*p <= '9')) p++;
   if (t == p) return(STRING_TYPE);
   if (*p == 0) return(INT_TYPE);
#ifndef FLOAT_TYPE
   return(STRING_TYPE);
#else
   /* now down to float case */
   if (*p == '.')
     {
	p++;
	while ((*p >= '0') && (*p <= '9')) p++;
     }
   if (*p == 0) return(FLOAT_TYPE);
   if ((*p != 'e') && (*p != 'E')) return(STRING_TYPE);
   p++;
   if (*p == '-') p++;
   while ((*p >= '0') && (*p <= '9')) p++;
   if (*p != 0) return(STRING_TYPE); else return(FLOAT_TYPE);
#endif
}

/* a literal */
int interp_push_number(char *t)
{
   int i;
   unsigned char stype;
   long value = 0;
#ifdef FLOAT_TYPE
   float x;
#endif

   stype = is_number(t);
   if (stype == STRING_TYPE) return(0);
   if (stype == INT_TYPE)
     {
	i = atoi(t);
	value = (long) i;
     }

#ifdef FLOAT_TYPE
   else if (stype == FLOAT_TYPE)
     {
	x = atof(t);
	value = *(long *) &x;
     }
#endif

   if (!Lang_Defining_Block && !Lang_Defining_Function)
     {
#ifdef FLOAT_TYPE
	if (stype == INT_TYPE)
	  {
#endif
	     lang_push_integer(i);
#ifdef FLOAT_TYPE
	  }
	else lang_push_float(x);
#endif
	return(1);
     }
   /* a literal */
   Lang_Object_Ptr->type = LANG_LITERAL | (stype << 8);
   Lang_Object_Ptr->value = value;
   Lang_Object_Ptr++;
   return(1);
}

/* only supports non negative integers, use 'chs' to make negative number */

void lang_check_space(void)
{
   int n;
   Lang_Object_Type *p;

   if (Lang_Interp_Stack_Ptr - Lang_Interp_Stack >= 3)
     {
	lang_doerror("Interpret stack overflow.");
	return;
     }

   if (Lang_Defining_Block)
     {
	n = (int) (Lang_Object_Ptr - Lang_Block_Body);
	if (n + 1 < Lang_BBody_Size) return;   /* extra for terminator */
	p = Lang_Block_Body;
     }
   else if (Lang_Defining_Function)
     {
	n = (int) (Lang_Object_Ptr - Lang_Function_Body);
	if (n + 1 < Lang_FBody_Size) return;
	p = Lang_Function_Body;
     }
   else return;

   /* enlarge the space by 2 objects */
   n += 2;
   if (NULL == (p = (Lang_Object_Type *) REALLOC(p, n * sizeof(Lang_Object_Type))))
     {
	lang_doerror("Lang: realloc failure.");
	return;
     }

   if (Lang_Defining_Block)
     {
	Lang_BBody_Size = n;
	n = (int) (Lang_Object_Ptr - Lang_Block_Body);
	Lang_Block_Body = p;
	Lang_Object_Ptr = p + n;
     }
   else
     {
	Lang_FBody_Size = n;
	n = (int) (Lang_Object_Ptr - Lang_Function_Body);
	Lang_Function_Body = p;
	Lang_Object_Ptr = p + n;
     }
}

int Lang_Defining_Variables = 0;

int lang_is_defined(char *what)
{
   Lang_Name_Type *n;
   compute_hash((unsigned char *) what);
   n = lang_locate_name_in_table(what, Lang_Intrinsic_Name_Table, Lang_Intrinsic_Name_Table + Name_Index[Hash], -1);
   if ((n != NULL) && (*n->name != 0)) 
     {
	/* Intrinsic-- return positive 1 if function, -1 if variable */
	return (n->obj.type & 0xFF) == LANG_INTRINSIC ? 1 : -1;
     }
   
   n = lang_locate_name_in_table(what, Lang_Name_Table, Lang_Name_Table, LANG_MAX_SYMBOLS);
   if ((n == NULL) || (*n->name == 0)) return(0);
   return (n->obj.type & 0xFF) == LANG_FUNCTION ? 2 : -2;
}

void lang_add_variable(char *name)
{
   Lang_Name_Type *table;
   long value;

   char *p, ch, *err = "Name Syntax";
   
   /* check syntax.  Allowed chars are: $!_?AB..Zab..z0-9 */
   ch = *name;
   if ((ch <= '9') && (ch >= '0')) 
     {
	lang_doerror(err);
	return;
     }

   p = name;
   while ((ch = *p++) != 0)
     {
	if ((ch >= 'a') && (ch <= 'z')) continue;
	if ((ch >= 'A') && (ch <= 'Z')) continue;
	if ((ch >= '0') && (ch <= '9')) continue;
	if ((ch == '_') || (ch == '$') || (ch == '!') || (ch == '?')) continue;
	lang_doerror(err);
	return;
     }
   
   if (strlen(name) > LANG_MAX_NAME_LEN)
     {
	lang_doerror("Name too long.");
	return;
     }

   if (Lang_Defining_Function)	       /* local variable */
     {
	compute_hash((unsigned char *) name);
	table = Lang_Local_Variable_Table;
	if (!Local_Variable_Number)
	  {
	     table = (Lang_Name_Type *) CALLOC(MAX_LOCAL_VARIABLES, sizeof(Lang_Name_Type));
	     if (NULL == table)
	       {
		  lang_doerror("Lang: calloc error.");
		  return;
	       }
	     Lang_Local_Variable_Table = table;
	  }
	strcpy(table[Local_Variable_Number].name + 1, name);
	*table[Local_Variable_Number].name = (char) Hash;
	table[Local_Variable_Number].obj.type = LANG_LVARIABLE;
	table[Local_Variable_Number].obj.value = Local_Variable_Number;
        Local_Variable_Number++;
     }
   else	if (!lang_is_defined(name))
     {
	if (0 == (value = lang_make_object())) return;
	lang_add_name(name, value, LANG_GVARIABLE);
     }
}

void interp_push_string(char *t)
{
   int len;

   /* strings come in with the quotes attached-- knock em off */
   if (*t == '"')
     {
	len = strlen(t) - 1;
	if (*(t + len) == '"') *(t + len) = 0;
	t++;
     }

   if (!Lang_Defining_Block && !Lang_Defining_Function)
     {
	lang_push_string(t);
	return;
     }

   /* a literal --- not to be freed */
   Lang_Object_Ptr->type = LANG_LITERAL | (STRING_TYPE << 8);
   if (NULL == (Lang_Object_Ptr->value = (long) lang_make_string((char *) t)))
     return;
   Lang_Object_Ptr++;
}

/* if an error occurs, discard current object, block, function, etc... */
void lang_restart(int localv)
{
   int save = Lang_Error;
   Lang_Error = UNKNOWN_ERROR;

   Lang_Break = Lang_Continue = Lang_Return = 0;
   while(Lang_Defining_Block)
     {
	lang_end_block();
     }

   if (Lang_Defining_Function)
     {
	if (Lang_Function_Body != NULL)
	  {
	     lang_define_function(NULL);
	     lang_free_branch(Lang_Function_Body);
	     FREE(Lang_Function_Body);
	  }
	if (Local_Variable_Number)
	  {
	     FREE(Lang_Local_Variable_Table);
	     Local_Variable_Number = 0;
	     Lang_Local_Variable_Table = NULL;
	  }
	Lang_Defining_Function = 0;
     }

   Lang_Error = save;
   /* --- warning--- I need to free things on the stack! */
   if (Lang_Error == STACK_OVERFLOW) Lang_Run_Stack_Depth = -1;
   
   Lang_Interp_Stack = Lang_Object_Ptr = Lang_Interp_Stack_Ptr = Lang_Interp_Stack_Static;
   /* This should be handled automatically */
   
   if (localv) Local_Variable_Frame = Local_Variable_Stack;
   Lang_Defining_Variables = 0;
}

void lang_compile(char *t)
{
   static int flag = 0;
   int d = 0;
   char ch = *t;
   
   lang_check_space();                 /* make sure there is space for this */
   
   
   if (!Lang_Error)
     {
	if (eqs(t,"{"))
	  {
	     lang_begin_block();
	     d = 1;
	  }
	else
	  {
	     d = try_directive(t, &flag);
	     if ((!flag && d) || (flag && !d)) Lang_Error = SYNTAX_ERROR;
	  }
	flag = 0;
     }

   if (Lang_Error || d);  /* null... */
   else if (Lang_Defining_Variables)
     {
	if (eqs(t, "]")) Lang_Defining_Variables = 0;
	else lang_add_variable(t);
     }
   else if (Lang_Defining_Function == -1) lang_define_function(t);
   else if (ch == '"') interp_push_string(t);
   else if ((ch == ':') && (Lang_Defining_Block))
     {
	Lang_Object_Ptr->type = LANG_LABEL;
	Lang_Object_Ptr->value = 0;
	Lang_Object_Ptr++;
     }

   else if ((Lang_Defining_Function == 1) && eqs(t,")"))
     {
	if (Lang_Defining_Block) lang_doerror("Function nesting illegal.");
	else Lang_Defining_Function = -1;
     }

   else if (eqs(t,"{"))
     {
	lang_begin_block();
	flag = 0;
     }

   else if (Lang_Defining_Block && eqs(t,"}"))
     {
	lang_end_block();
	flag = 1;
     }

   else if (eqs(t,"("))	lang_begin_function();

   else if (eqs(t, "[")) Lang_Defining_Variables = 1;
   else if (lang_try_binary(t));
   else if (lang_try_unary(t));

   /* note that order here is important */
   else if ((ch <= '9') && interp_push_number(t));
   else if ((ch <= '=') && interp_variable_eqs(t));
   else if (lang_exec(t));
   else Lang_Error = SYNTAX_ERROR;

   if (Lang_Error) lang_restart(0), flag = 0;
}


#ifdef Future_Slang

void pretty_print_object(Lang_Object_Type *x)
{
   if ((x->type >> 8) == STRING_TYPE)
     {
	fprintf(stdout,"\"%s\"    ", (char *) x->value);
     }
   else if ((x->type >> 8)  == INT_TYPE)
       fprintf(stdout,"%ld    ", x->value);
#ifdef FLOAT_TYPE
   else
     if ((x->type >> 8)  == FLOAT_TYPE)
     fprintf(stdout,"%g    ", (float) *(float *) &x->value);
#endif
}

void print_stack(void)
{
   int i, depth;

   depth = Lang_Run_Stack_Depth;
   pretty_print_object(&Lang_Run_Stack[depth]);
   fputs("  (",stdout);
   for (i = depth - 1; i >= 0; i--) pretty_print_object(&Lang_Run_Stack[i]);
   fputs(")",stdout);
   fflush(stdout);
}
#endif

/* optimize strchr */
#define STRCHR(x, y) ((y >= 'a') && (y <= 'z') ? NULL : strchr(x, y))

int extract_token(char **linep, char *word)
{
   char ch, *line;
   char *special = " \t{}[];():";
   int num, i = 0;
   int string;

    line = *linep;

    /* skip white space */
    while(ch = *line++, (ch == ' ') || (ch == '\t'));

    if ((!ch) || (ch == '\n'))
      {
	 *linep = line;
	 return(0);
      }

   *word++ = ch;
   if (ch == '"') string = 1; else string = 0;
   if (ch == '\'')
     {
	if (((ch = *line++) != 0) && (*line++ == '\'')) 
	  {
	     i = (int) ((unsigned char) ch);
	     --word;
	     sprintf(word, "%d", (int) ((unsigned char) ch));
	     word += 4;
	  }
	else Lang_Error = SYNTAX_ERROR;
     }
   else
    if (NULL == STRCHR(special, ch))
      while(ch = *line++, (ch != '\n') && (ch != 0) &&  (ch != '"'))
	{
	   if (string)
	     {
		if (ch == '\\')
		  {
		     ch = *line++;
		     if ((ch == 0) || (ch == '\n')) break;
		     switch(ch)
		       {
			  case 'n': ch = '\n'; break;
			  case 'r': ch = '\r'; break;
			  case 't': ch = '\t'; break;
			  case 'e': ch = '\033'; break;  /* escape */
			  case 'b': ch = '\007'; break;  /* bell */
			default:
			  /* \xyz is ascii value xyz! */
			  if ((ch < '0') || (ch > '2')) break;
			  num = 0;
			  for (i = 0; i < 3; i++)
			    {
			       num = num * 10 + (int) (ch - '0');
			       ch = *line++;
			       if ((ch < '0') || (ch > '9')) break;
			    }
			  line--;
			  /* should check for parse error here! */
			  ch = num;
		       }
		  }
	     }
	   else if (NULL != STRCHR(special,ch))
	     {
		line--;
		break;
	     }

           *word++ = ch;
	}

   if ((!ch) || (ch == '\n')) line--;
   if ((ch == '"') && string) *word++ = '"'; else if (string) Lang_Error = SYNTAX_ERROR;
   *word = 0;
   *linep = line;
   return(1);
}

/* interprets line-- returns offset of last part of line evaluated */
char *interpret(line)
char *line;
{
   char token[255];
   char *ret;

   while(ret = line, extract_token(&line,token))
     {
	if (Lang_Error) break;
	if (*token == ';') break;
	lang_compile(token);
	if (Lang_Error) break;
	/* puts(token); */
     }
   if (Lang_Error) lang_doerror(NULL);
   return(ret);
}

#ifdef pre_jed090
/* Allow easy addition of new intrinsic functions */
void add_intrinsic(char *name, long f, unsigned char out, int argc)
{
   unsigned short type;
   unsigned char stype;

   stype = out | (argc << 4);
   type = LANG_INTRINSIC | (stype << 8);
   lang_add_name(name, f, type);
}

void add_variable(char *name, long addr, int vtype, int ro)
{
   unsigned short type;

   if (ro) type = LANG_RVARIABLE; else type = LANG_IVARIABLE;
   type = type | (vtype << 8);
   lang_add_name(name, addr, type);
}
#endif

long *lang_add_array(char *name, long* addr, int dim, int d0, int d1, int d2, char *t)
{
   unsigned short type;
   long a = (long) addr;

   /* This is not really true since if addr is not (long) NULL, the space is
      set aside by the compiler.  Here we assume it is dynamic.  The upshot
      is that the user might pass it to free_array which will cause problems. */

   addr = (long *) lang_create_array((long *) &a, dim, d0, d1, d2, t);
   type = LANG_IVARIABLE | (ARRAY_TYPE << 8);
   lang_add_name(name, (long) addr, type);
   return((long *) a);
}

static void lang_getenv_cmd(char *s)
{
   char *t;
   if (NULL == (t = getenv(s))) t = "";
   lang_push_string(t);
}

int lang_apropos1(char *s, Lang_Name_Type *table, int max)
{
   int all = 0, n = 0;
   char *nm;
   Lang_Object_Type obj;

   if (*s == 0) all = 1;

   while(max && (nm = table->name, *nm != 0))
     {
	nm++;  /* lose hash */
	if ((*nm != 1) && (all || (NULL != strstr(nm, s))))
	  {
	     n++;
	     /* since string is static, make it literal */
	     obj.type = LANG_LITERAL | (STRING_TYPE << 8);
	     obj.value = (long) nm;
	     lang_push(&obj);
	     if (Lang_Error) return(1);
	  }
	table++;
	max--;
     }
   return n;
}

void lang_apropos(char *s)
{
   int n;
   
   n = lang_apropos1(s, Lang_Intrinsic_Name_Table, -1);
   n += lang_apropos1(s, Lang_Name_Table, LANG_MAX_SYMBOLS);
   lang_push_integer(n);
}

void init_lang()
{
   Lang_Name_Type *e0, *entry = Lang_Intrinsic_Name_Table;
   char *name;
      
   lang_add_variable(SLANG_SYSTEM_NAME);
   
   /* compute hash values */
   
   e0 = entry;
   while (name = entry->name, *name != 0)
     {
	compute_hash((unsigned char *) (name + 1));
	*name = (char) Hash;
	if (Name_Index[(unsigned char) Hash] == -1)
	  {
	     Name_Index[(unsigned char) Hash] = (int) (entry - e0);
	  }
	entry++;
     }
}

/* probably more useful to have a argc, argv thing */
int lang_run_hooks(char *hook, char *opt1, char *opt2)
{
   int ret = 0;

   if (!lang_is_defined(hook)) return(0);
   if (opt1 != NULL) lang_push_string(opt1);
   if (opt2 != NULL) lang_push_string(opt2);
   if (!Lang_Error) ret = lang_execute_function(hook);
   return (ret && !Lang_Error);
}


/*  name table */

Lang_Name_Type Lang_Intrinsic_Name_Table[] =
{
   MAKE_INTRINSIC(".autoload",  lang_autoload, VOID_TYPE, 2),
   MAKE_INTRINSIC(".pop",  do_pop, VOID_TYPE, 0),
   MAKE_INTRINSIC(".strcmp",  do_strcmp, VOID_TYPE, 0),
   MAKE_INTRINSIC(".strcat",  do_strcat, VOID_TYPE, 0),
   MAKE_INTRINSIC(".strlen",  do_strlen, VOID_TYPE, 0),
   MAKE_INTRINSIC(".defined?",  lang_is_defined, INT_TYPE, 1),
   /* returns 0 if arg is not defined.
      return +1 if arg is an intrinsic function 
	     +2 if user defined function
	     -1 if intrinsic variable
	     -2 if user defined variable */
   MAKE_INTRINSIC(".string",  do_string, VOID_TYPE, 0),
   MAKE_INTRINSIC(".getenv",  lang_getenv_cmd, VOID_TYPE, 1),
   MAKE_INTRINSIC(".evalfile",  lang_load_file, INT_TYPE, 1),
   MAKE_INTRINSIC(".char",  do_char, VOID_TYPE, 0),
   MAKE_INTRINSIC(".eval",  interpret, VOID_TYPE, 1),
   MAKE_INTRINSIC(".dup",  do_dup, VOID_TYPE, 0),
   MAKE_INTRINSIC(".substr",  do_substr, VOID_TYPE, 0),
   /* Syntax: "string" n len substr
       returns a substring with length len of string beginning at position n.
     */
   MAKE_INTRINSIC(".integer",  do_integer, VOID_TYPE, 0),
   /* Convert from a string representation to integer.  For example,
      "1234" float returns 1234 to stack. */
   MAKE_INTRINSIC(".is_substr",  do_issubstr, VOID_TYPE, 0),
   /* Syntax: "a" "b" is_substr
      returns the position of "b" in "a".  If "b" does not occur in "a"
      it returns 0--- the first position is 1 */
   MAKE_INTRINSIC(".strsub",  do_strsub, VOID_TYPE, 0),
   /* Syntax:  "string"  n ascii_value strsub
     This forces string to have a char who asciii value is ascii_val at
     the nth position.  The first character in the string is at position
     1. */
#ifdef FLOAT_TYPE
   MAKE_INTRINSIC(".float",  do_float, VOID_TYPE, 0),
   /* Convert from a string representation to floating point.  For example,
      "12.34" float returns 12.34 to stack. */
#endif

   MAKE_VARIABLE(".TRACEBACK", &Lang_Traceback, INT_TYPE, 0),
   /* If non-zero, dump S-Lang tracback on error. */
   
#include "intrin.ci"

   /* these are rarely ever referred to so make them last. */
   MAKE_INTRINSIC(".system",  system, INT_TYPE, 1),
   MAKE_INTRINSIC(".slapropos",  lang_apropos, VOID_TYPE, 1),
   /* array ops: */
   MAKE_INTRINSIC(".create_array",  create_array, VOID_TYPE, 0),
   /* Syntax: "type" i_1 i_2 ... i_dim dim create_array
      returns ARRAY object to stack with dimension dim.  i_n is
      an integer which specifies the maximum size of array in direction n.
      "type" is a control string which specifies the type of the array.
      Types are:  "s" : array of strings
                  "f" : array of floats
                  "i" : array of integers
      At this point, dim cannot be larger than 3.
      Also note that space is dynamically allocated for the array and that
      copies of the array are NEVER put on the stack.  Rather, references to
      the array are put on the stack.  When the array is no longer needed, it
      must be freed with 'free_array' */
   MAKE_INTRINSIC(".free_array",  free_array, VOID_TYPE, 0),
   /* Syntax: ARRAY free_array
      Frees up the space which array occupies.  All reference to this space
      will now be meaningless. */
   MAKE_INTRINSIC(".aget",  array_getelem, VOID_TYPE, 0),
   /* Syntax: i j ... k  ARRAY aget
      returns ARRAY[i][j]...[k] */
   MAKE_INTRINSIC(".aput",  array_putelem, VOID_TYPE, 0),
   /* Syntax: x i j ... k  ARRAY put
      sets ARRAY[i][j]...[k] = x */

   MAKE_INTRINSIC(".strncmp",  do_strncmp, VOID_TYPE, 0),
   MAKE_INTRINSIC(".strlow", do_strlow, VOID_TYPE, 0),
   /* Takes a string off the stack a replaces it with all characters
      in lowercase. */
   MAKE_INTRINSIC(".strup", do_strup, VOID_TYPE, 0),
   /* Takes a string off the stack a replaces it with all characters
      in uppercase. */
   MAKE_INTRINSIC(".isdigit",  do_isdigit, INT_TYPE, 1),
   MAKE_INTRINSIC(".strtrim", do_strtrim, VOID_TYPE, 0),
   /* Trims leading and trailing whitespace from a string.  WHitespace
      is defined to be spaces, tabs, and newline chars. */

   MAKE_INTRINSIC(".int",  do_int, VOID_TYPE, 0),
   /* returns ascii value of the first character of a string. */
   MAKE_INTRINSIC(".array_sort", lang_array_sort, VOID_TYPE, 1),
   /* Requires an array on the stack as well as a function name to call 
    for the comparison.  The array to be placed on the stack is the
    array to be sorted.  The routine returns an integer index array which 
    indicates the result of the sort.  The first array is unchanged. */
       
#define THE_END MAKE_INTRINSIC
   THE_END("", 0, 0, 0)
};

